"""
{% timedelta cart.updated cart.created format="" %}
{{ cart.created|timesince:cart.updated }}
{% human_timedelta cart.updated cart.created format="" %}
"""

from datetime import timedelta

from django import template
from django.urls import reverse
from django.utils.translation import ngettext

register = template.Library()


@register.simple_tag(takes_context=False)
def str_merge(*targs, **kwargs):
    """Join many separate strings into a single string

    {% str_merge "foo" var ".baz" as my_str %}
    {{ my_str }}
    """
    return "".join(map(str, targs))



"""
Unit tests for updated_params template tag.
"""

import unittest

# from django.test import RequestFactory
# from django.template import Context
from trim.templatetags import updated_params


class SimGet:

    def copy(self):
        # Simulate a copy method for GET parameters
        return self

    def update(self, *a, **kwargs):
        # Simulate an update method for GET parameters
        pass

    def urlencode(self):
        # Simulate urlencode for testing
        return "simulated urlencode"


class SimContext:
    #   self.context = { 'request':{'GET':{}}}#Context({'request': self.request})

    def __init__(self):
        self.GET = SimGet()


class TestUpdatedParamsTag(unittest.TestCase):
    def setUp(self):
        pass
        # Setup a mock request and context for template tag
        # self.factory = RequestFactory()
        # self.request = self.factory.get('/?foo=bar')
        self.context = {"request": SimContext()}  # Context({'request': self.request})

    def test_updated_params_basic(self):
        """Test that updated_params returns updated query string."""
        # Setup
        kwargs = {"page": 2, "foo": "baz"}
        # Expected: Query string should be updated with new params
        # Result
        result = updated_params.updated_params(self.context, **kwargs)
        # Check if the result matches expected query string
        self.assertEqual(result, "simulated urlencode")

    def test_updated_params_no_kwargs(self):
        """Test updated_params with no additional kwargs."""
        # Setup
        kwargs = {}
        # Expected: Query string should remain unchanged
        # Result
        result = updated_params.updated_params(self.context, **kwargs)
        # TODO: Add assertions for unchanged query string
        pass



from trim import views

from .. import dispatch, forms, models

# __all__ = ['EmailInviteCreateView']


class EmailInviteCreateView(views.CreateView):
    model = models.EmailInvite
    fields = (
        "email_address",
        # 'user',
    )

    def get_initial(self):
        return {"user": self.request.user}

    def form_valid(self, form):
        form.instance.user = self.request.user
        self.object = form.save()
        dispatch.send_invite.send(sender=self.__class__, object=self.object)
        return super().form_valid(form)

    def get_success_url(self):
        return views.reverse("account:profile")


class EmailInviteListView(views.ListView):
    model = models.EmailInvite

    def get_queryset(self):
        return super().get_queryset().filter(user=self.request.user)



from django import template
from django.template import Context
from django.template.base import token_kwargs

from . import quickforms


def inject_node(parser):
    T = parser.tokens[0].__class__
    tr = "{{ slotplot }}"
    parser.prepend_token(T(parser.tokens[0].token_type, tr))


register = template.Library()


def parse_until(parser, token, ends, delete_first=True):
    nodelist = parser.parse(ends)
    if delete_first:
        parser.delete_first_token()
    # inject_node(parser)
    splits = token.split_contents()[1:]
    extra = {}
    for i, word in enumerate(splits):
        if word == "with":
            extra = token_kwargs(splits[i + 1 :], parser)
    return nodelist, splits, extra


@register.tag(name="wrap")
def do_wrap(parser, token):
    """Wrap the contents given to the nodem with another template given through
    the tag token:

        {% wrap "stocks/wrap_form.html" with button_text="Next"  action="/fo/bar" %}
            <ul>
            {% for row in form %}
                <li>{{ row }}</li>
            {% endfor %}
            </ul>
        {% endwrap %}

    The template "wrap_forms.html" should refer to the `wrap.content` and any
    other variables from the `with` or parent contexts.

        <form method="post"
          enctype="multipart/form-data"
          {% if action %}action="{{action}}"{% endif %}>

            {{ wrap.content }}
            <input class='btn'
                type="submit"
                value="{% firstof button_text 'Confirm' %}">
            {% csrf_token %}
        </form>
    """
    nodelist, splits, extra = parse_until(parser, token, ("endwrap",))
    return WrappedContentNode(nodelist, splits, **extra)


class WrappedContentNode(template.Node):
    def __init__(self, nodelist, tokens, *a, **kw):
        self.nodelist = nodelist
        self.token_template_name = template.Variable(tokens[0])
        self.extra_context = kw

    def render(self, context):
        print("rendering wrap")
        template_name = self.token_template_name.resolve(context)
        t = context.template.engine.get_template(template_name)
        content = self.nodelist.render(context)

        # we extend the original context; allowing this unit to be overwitten
        # if required.
        wrap = {"content": content, "template_name": template_name}
        #  Dynamic key for the root context, static key for the 'wrap' object.
        content_key = "content"
        sub_ctx = {"wrap": wrap, content_key: content}
        values = {key: val.resolve(context) for key, val in self.extra_context.items()}
        with context.push(**sub_ctx, **values):
            # Context({'source': content}, autoescape=context.autoescape)
            return t.render(context)


@register.tag(name="slot")
def do_slot(parser, token):
    nodelist, splits, extra = parse_until(parser, token, ("endslot",))
    return SlotNode(nodelist, splits, **extra)


class SlotNode(template.Node):
    def __init__(self, nodelist, tokens, *a, **kw):
        self.nodelist = nodelist
        # self.token_template_name = template.Variable(tokens[0])
        self.extra_context = kw

    def render(self, context):
        # template_name = self.token_template_name.resolve(context)
        # t = context.template.engine.get_template(template_name)
        print("rendering slot")
        content = self.nodelist.render(context)
        # we extend the original context; allowing this unit to e overbwitten
        # if required.
        wrap = {
            "content": content,
        }
        #  Dynamic key for the root context, static key for the 'wrap' object.
        content_key = "content"
        sub_ctx = {}  # = {'wrap':wrap, content_key:content}
        values = {key: val.resolve(context) for key, val in self.extra_context.items()}
        with context.push(**sub_ctx, **values):
            # Context({'source': content}, autoescape=context.autoescape)
            return self.nodelist.render(context)


# @register.simple_tag(takes_context=True)
# def info_form(context, product_id):
#     return forms.ProductQuestionForm(initial={
#             'product_id':product_id,
#         })
#
# @register.inclusion_tag('dummy.html')
# def wrap(template='default.html'):
#     return {'template': template}



import json
import os
import uuid
from pathlib import Path

from django.conf import settings
from django.core.files.storage import FileSystemStorage, default_storage
from django.http import JsonResponse
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt, csrf_protect

from ..forms.upload import FileChunkForm, FilesForm, MergeConfirmForm
from ..merge import FileExists, recombine
from .base import FormView, TemplateView

HERE = Path(__file__).parent
ROOT = HERE.parent.parent.parent
UPLOADS = ROOT / "uploads"

# fs = FileSystemStorage(location=UPLOADS)

cache = {}


def get_cache():
    return cache


def unlink_dir_files(dir_path):
    res = ()
    for root, dirs, files in os.walk(dir_path):
        for f in files:
            os.unlink(os.path.join(root, f))
            # res += (Path(f).relative_to(dir_path),)
            res += (f,)
        for d in dirs:
            shutil.rmtree(os.path.join(root, d))
            # res += (Path(d).relative_to(dir_path),)
            res += (d,)

    c = 0
    for root, dirs, files in os.walk(dir_path):
        c += len(files)
        c += len(dirs)

    return {
        "files": res,
        "old_count": len(res),
        "new_count": c,
    }


def verify_file(asset):
    res = {"size": asset["output"]["size"] == asset["bytesize"]}

    return res


class AssetMixin(object):

    upload_dir_settings_key = "CHUNK_UPLOAD_DIR"

    def get_uuid(self):
        return self.kwargs["uuid"]

    def get_asset(self):
        return cache[self.get_uuid()]

    def ensure_dir(self, location):
        fullpath = location
        print("Destination", fullpath)

        if fullpath.exists() is False:
            fullpath.mkdir(parents=True)
            print(fullpath)
        return fullpath

    def get_upload_dir(self):
        # HERE = Path(__file__).parent
        # ROOT = HERE.parent.parent.parent
        # UPLOADS = ROOT / "uploads"
        return getattr(settings, self.upload_dir_settings_key)

    def get_fs(self):
        uploads = self.get_upload_dir()
        return FileSystemStorage(location=uploads)

    def get_current_username(self):
        username = self.request.user.username
        if len(username) == 0 and self.request.user.is_anonymous:
            username = "anonymous"
        return username

    def get_parts_dir(self):
        file_uuid = self.get_uuid()
        info = cache[file_uuid]
        name = info["internal_name"]
        suffix = info["suffix"]
        username = self.get_current_username()
        make_name = Path(username) / file_uuid
        return make_name


class UploadAssetView(FormView, AssetMixin):
    """An upload file form view, with two forms."""

    form_class = FilesForm
    template_name = "trim/upload/upload_view.html"

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs["file_chunk_form"] = FileChunkForm
        return kwargs

    def form_valid(self, form):
        """Return an ajax response of content
        for the incoming chunks
        """
        data = form.cleaned_data
        file_uuid = str(uuid.uuid4())
        filename = data["filename"]
        if len(filename) == 0:
            filename = "foo.unknown"

        suffix = Path(data["filepath"]).suffix
        name = Path(data["filename"]).stem
        cache = get_cache()
        cache[file_uuid] = {
            "filetype": data["filetype"],
            "bytesize": data["byte_size"],
            "filepath": data["filepath"],
            "suffix": suffix,
            "internal_name": name,
            "done": False,
        }

        cache[file_uuid]["filename"] = filename

        return JsonResponse(
            {
                "ok": True,
                "file_uuid": file_uuid,
            }
        )


class UploadAssetSuccessView(TemplateView, AssetMixin):
    template_name = "corsa/upload_success.html"

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs["asset"] = self.get_asset()
        return kwargs


@method_decorator(csrf_exempt, name="dispatch")
class UploadChunkView(FormView, AssetMixin):
    form_class = FileChunkForm
    template_name = "trim/upload/upload_view.html"

    def ensure_fullpath(self, filename):
        fs = self.get_fs()
        location = Path(fs.location)
        fullpath = location / filename

        print("Location", location)
        print("Destination", fullpath)

        if fullpath.parent.exists() is False:
            if location.exists():
                # ensure the parent is set correctly.
                v = fullpath.relative_to(location)
                fullpath.parent.mkdir(parents=True)
                print(v)
        return fullpath

    def form_valid(self, form):
        data = form.cleaned_data
        # returned from the upload asset initial view.
        file_uuid = data["file_uuid"]
        info = cache.get(file_uuid)

        # The chunk of file
        filepart = data["filepart"]
        # The int chunk index
        index = data["chunk_index"]
        # The mapped name given to the form before the main upload
        name = info["internal_name"]
        suffix = info["suffix"]

        username = self.get_current_username()

        # The 'part' defines an extended suffix, to flag this is a partial
        # file.
        part = f".part_{index}"
        filename = f"{name}{suffix}{part}"  # received file name
        make_name = Path(username) / file_uuid / filename
        # fs.location is enforced.
        store_path = self.ensure_fullpath(make_name)

        ok = self.write_file(filepart, store_path)

        return JsonResponse({"ok": True})

    def write_file(self, file, store_path, storage=None):
        print("Writing", store_path)
        storage = storage or self.get_fs()
        # with fs.open(store_path, 'wb+') as stream:
        with storage.open(store_path, "wb+") as stream:
            for chunk in file.chunks():
                stream.write(chunk)
        return store_path.exists()


class MergeAssetView(FormView, AssetMixin):

    form_class = MergeConfirmForm
    # template_name = "corsa/form.html"
    template_name = "trim/file/merge_view.html"

    def get_success_url(self):
        args = (self.get_uuid(),)
        return reverse("corsa:upload_success", args=args)

    def get_out_dir(self):
        fs = self.get_fs()
        out_path = Path(fs.location) / "finished" / self.get_current_username()
        return out_path

    def form_valid(self, form):
        # Confirm post to merge the attached asset
        data = form.cleaned_data
        accept = data["confirm_merge"]
        if not accept:
            return super().form_valid(form)

        asset = self.get_asset()

        out_path = self.get_out_dir()
        self.ensure_dir(out_path)

        fs = self.get_fs()
        dir_path = Path(fs.location) / self.get_parts_dir()
        asset["input"] = {
            "path": dir_path,
        }

        if dir_path.exists():
            self.perform(asset, dir_path, out_path)
            if data["delete_cache"]:
                self.delete_cache(asset)
            return super().form_valid(form)

        print("Bad Path.")
        return super().form_valid(form)

    def delete_cache(self, asset):
        # delete the merge parts.
        print("Delete")
        p = Path(asset["input"]["path"])
        if p.exists() and p.is_dir():
            # scrub it
            print("Delete", p)
            deleted = unlink_dir_files(p)
            asset["deletion"] = deleted
            receipt = p / "delete-receipt.json"
            receipt.write_text(json.dumps(deleted, indent=4))
        return p.exists() is False

    def perform(self, asset, dir_path, out_path):
        try:
            output_path = recombine(dir_path, out_path)
        except FileExists as err:
            output_path = err.args[0]

        output_path = Path(output_path)
        fs = self.get_fs()

        asset["output"] = {
            "path": output_path.relative_to(fs.location).as_posix(),
            "exists": output_path.exists(),
            "size": os.path.getsize(output_path),
            "sub_path": output_path.relative_to(out_path).as_posix(),
        }

        asset["output_path"] = Path(asset["output"]["path"]).as_posix()
        asset["done"] = True
        asset["verification"] = verify_file(asset)

        return asset



from django.db.models import Model

from .models.auto import hook_init_model_mixins


def str_printer(self, alts=None):
    """Acting as a __str__ replacement. Provide a list of format strings
    as `alts`, with a natural default to '_trim_string'.
    If a function exists matching the string format [get]_trim_string, the function is called.
    """
    apply_label = getattr(self, "_trim_props_label", True)
    self_prop_str = "{self.%(prop)s}"
    default_prop_format = (
        f'%(prop)s="{self_prop_str}"' if apply_label else self_prop_str
    )
    prop_format = getattr(self, "_trim_props_format", default_prop_format)
    alts = (alts or ()) + ("_trim_string",)
    format_str = None

    # Grab the first (best) string format - use a get method if it exists.
    #   self: _trim_string_repr, _trim_string, ...
    for s in alts:

        if hasattr(self, f"get{s}"):
            # self.get_trim_string()
            format_str = getattr(self, f"get{s}")()
            break

        if hasattr(self, s):
            format_str = getattr(self, s)
            break

    if format_str is None:
        # _trim_string or other alt props are not applied.
        if hasattr(self, "_trim_props"):
            props = self._trim_props
            if isinstance(props, (tuple, list)) is False:
                props = (props,)
                # If only one string, then reduce the printout "field=X",
                # to just "X"
                prop_format = self_prop_str

            format_str = ", ".join(prop_format % {"prop": x} for x in props)

    return format_str.format(self=self)


def repr_printer(self):
    cl = self.__class__.__name__
    alts = ("_trim_string_repr",)
    return f"<{cl}({self.pk}) '{str_printer(self, alts)}'>"


def model_pre_init(sender, args, kwargs, **kw):
    hook_init_model_mixins(sender)

    if hasattr(sender, "_trim_string") or hasattr(sender, "_trim_props"):
        # print('model_pre_init', sender, kw)
        if sender.__str__ == Model.__str__:
            sender.__str__ = str_printer

        if sender.__repr__ == Model.__repr__:
            sender.__repr__ = repr_printer

        # sender.__repr__ == Model.__repr__



import argparse
import json
import os
import textwrap
from collections import defaultdict
from pathlib import Path

import appdirs

register = {"functions": defaultdict(tuple)}
# # create the parser for the "a" command
# parser_a = subparsers.add_parser('add', help='a help', aliases=['a'])
# parser_a.add_argument('bar', type=int, help='bar help')


class Help:
    primary = "Application primary help"
    scripts = "Work with the loaded trim scripts"
    add = "Add a script to the loadout"
    add_filename = "Provide the file for loading."


class SubHelpFormatter(argparse.HelpFormatter):
    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self._indent_increment = 2
        self._level += 2


def get_subactions(parser):
    return [
        action
        for action in parser._actions
        if isinstance(action, argparse._SubParsersAction)
    ]


def print_help(parser, less=False):
    # retrieve subparsers from parser
    subparsers_actions = get_subactions(parser)
    # there will probably only be one subparser_action,
    # but better safe than sorry
    for subparsers_action in subparsers_actions:
        # get all subparsers and print help
        for choice, subparser in subparsers_action.choices.items():
            print_sub_help((choice,), subparser, less=less)


DEFAULT_PREFIX = "-"


def print_sub_help(choices, subparser, depth=0, add_spaces=0, prefix=None, less=False):
    d = 1
    dashes = (DEFAULT_PREFIX if prefix is None else prefix) * (depth + d)
    spaces = " " * (depth + add_spaces)

    print("{}{} $ APP {}".format(spaces, dashes, " ".join(choices)))

    subparser.formatter_class = SubHelpFormatter

    if less is False:
        vv = subparser.format_help()
        spaces = " " * (depth + d + 3)
        vv = textwrap.indent(vv, spaces)
        print(vv)

    subparsers_actions = get_subactions(subparser)

    for subparsers_action in subparsers_actions:
        # get all subparsers and print help
        for choice, _subparser in subparsers_action.choices.items():
            print_sub_help(
                choices + (choice,),
                _subparser,
                depth=depth + 1 if less is False else d,
                less=less,
                prefix=prefix or "",
            )


class ConfigMixin(object):

    def get_conf(self):
        """Collect a data object from JSON representation as a persistent user
        config file.

            data = self.get_conf()
            data['change'] = 1
            self.write_conf_data(data)
        """
        # info_point = appdirs.user_data_dir('v1', 'django-trim')
        # # look at base location
        # init_path = Path(info_point) / 'trim-config.json'
        conf_path = self.get_conf_path()

        data = {}
        if conf_path.exists():
            try:
                data = json.loads(conf_path.read_text())
            except json.decoder.JSONDecodeError:
                print("The pointer config file is not JSON. Failing early.")
        return data

    def get_conf_path(self):
        info_point = appdirs.user_data_dir("v1", "django-trim")
        # look at base location
        init_path = Path(info_point) / "trim-config.json"
        conf_path = init_path

        if conf_path.exists():
            # load as json, test for pointer.
            print("Loading path.", conf_path)
            data = json.loads(conf_path.read_text())
            pointer = data.get("pointer", None)
            if pointer is not None:
                print("Pointer path.", pointer)
                conf_path = Path(pointer)
        return conf_path

    def write_conf_data(self, data, path=None):
        """Write a data object as JSON representation to the persistent user
        config file.

            data = self.get_conf()
            data['change'] = 1
            self.write_conf_data(data)
        """
        path = path or self.get_conf_path()
        jd = json.dumps(data, indent=4)
        try:
            path.write_text(jd)
        except FileNotFoundError as err:
            if path.parent.exists() is False:
                os.makedirs(path.parent)
            path.write_text(jd)


class AppFunction(ConfigMixin):
    """The AppFunction loadout helps to build attachments to the argparser.
    Apply this class to the AppActions app_functions to insanstiate during a prep.

    The class is ran against the AppActions to manipulate the argparser options.
    For convenience a few functions abstract the cli argument. The 'prep' function
    assumes a `hook` and `setup_args` function exist

       class ScriptAdd(AppFunction):
           # parent_name = 'scripts'
           parent_name = None
           name = 'add'
           help = Help.add

           def setup_args(self, parser):
               parser.add_argument('filename', type=Path, help=Help.add_filename)
               parser.add_argument('--name', type=str, default=None, )

           def hook(self, parsed, *args, **kwargs):
               print('hook', parsed)
               data = self.get_conf()
               name = parsed.name or parsed.filename.stem
               data[name] = parsed.filename.absolute().as_posix()
               self.write_conf_data(data)

               return 'res from hook'

    If the parent name is none, the function appends as a parser. If the parent
    name exists, the function appends as a subparser of the positioned parent.

       class Script(AppFunction):
           name = 'scripts'
           help = Help.scripts


       class ScriptAdd(AppFunction):
           parent_name = 'scripts'
           name = 'add'
           help = Help.add

    """

    # parent_name = 'scripts'
    parent_name = None
    name = None
    help = None
    arguments = None
    auto_register = True
    register_name = "default"

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        # print('Register', cls)
        if cls.auto_register is True:
            register["functions"][cls.register_name] += (cls,)

    def prep(self, app):
        """Prepare the tool with the pre-baked loadouts such as the help and
        scripts.
        """
        parser = self.hook_parser(app)
        for A in self.arguments or ():
            A().prep(app)
        if parser:
            self.setup_args(parser)
        return parser

    def hook_parser(self, app):
        pname = self.parent_name
        cname = self.name or self.__class__.__name__.lower()
        mypath = f"{app.primary_name}.{pname}"
        if pname is not None:
            return app.add_sub_to(mypath, self.hook, cname, help=self.help)
        return app.add_sub(self.hook, cname, help=self.help)

    def setup_args(self, parser):
        pass

    def hook(self, parsed, *args, **kwargs):
        pass


class AppArgument(AppFunction):
    """Append an agument in the form of a switch to an existing parser.


    class ScriptsAddFilenameArg(AppArgument):

        target ='scripts.add'

        def setup_args(self, parser):
            parser.add_argument('filename', type=Path)
            parser.add_argument('-n', '--name')

    AppArguments do not automatically register to the load lib as they
    may optionally append to a parent manually.
    """

    target = None
    # auto_register = False
    register_name = "default"

    def hook_parser(self, app):
        parser = self.get_parser(app, self.target)
        return parser

    def get_parser(self, app, name=None):
        try:
            r = app.get_parser(name)
            return r
        except NoPosition:
            pass


class NoPosition(Exception):
    pass


import argparse


class AppActions(ConfigMixin):
    prog_name = "PROG"
    primary_name = "primary"
    register_name = "default"

    def __init__(self):
        self._subparsers = {}
        self.args = None
        self.parser = None
        self.positions = {}
        self.primary_init_kwargs = {}
        self.app_functions = ()
        self.setup()

    def setup(self):
        pass

    def prep(self, app=None):
        """Prepare the tool with the pre-baked loadouts such as the help and
        scripts.
        """
        # scripts_name = 'scripts'
        # self.add_sub(self.scripts_func, scripts_name, help=Help.scripts)
        reg = self.get_register_function()
        for AppFuncClass in self.app_functions + reg:
            AppFuncClass().prep(self)

    def get_register_function(self):
        return register["functions"][self.register_name]

    def get_subparser(self, parser=None, **kwargs):
        """Return the 'subparser' of a given parent parser;
        such as "bar for "$foo bar". If the parent parser is not given, use
        the primary parser.
        """

        ## If add is False, the default (standard) cli printout occurs,
        # if true, (default), the internal printout occurs.
        ## By default it's on to allow both the default, and the optional
        ### $ appname [nothing]
        # kwargs.setdefault('add_help', False)
        parser = parser or self.get_primary_parser(**kwargs)

        if self._subparsers.get(parser) is None:
            subparsers = parser.add_subparsers(help="sub-command help")
            subparsers._parent = parser._name
            self._subparsers[parser] = subparsers
        return self._subparsers[parser]

    def add_sub(self, func, name, parser=None, **kwargs):
        """Add a subparser function as the "name" to a parent parser.
        If the parent parser is not given, use the primary.

        This function applied the default function and stored the positio for
        later rebinding.

            add_parser = self.add_sub(self.scripts_func, 'scripts', help=Help.scripts)
            #positions[primary.scripts] == add_parser
        """
        # create the parser for the "b" command
        subparsers = self.get_subparser(parser)
        parser_b = subparsers.add_parser(name, **kwargs)
        parser_b._name = name
        # parser_b.add_argument('--baz', choices='XYZ', help='baz help')
        parser_b.set_defaults(func=func)
        position = f"{subparsers._parent}.{name}"
        self.positions[position] = parser_b
        return parser_b

    def add_sub_to(self, position_name, func, name, **kwargs):
        """Simplify the creation of a sub loader to an expected position:

        self.add_sub(self.scripts_func, 'add', 'primary.scripts', help=Help.scripts)
        """
        return self.add_sub(func, name, self.positions[position_name], **kwargs)

    def default_caller(self, args):
        print_help(self.get_primary_parser(), less=True)
        return "res from default_caller"

    def get_primary_parser(self, **kwargs):
        """Return the primary parser for the program. If the internal self.parser
        is undefined, create a new 'primary' and return it.
        """
        if self.parser is None:
            # kwargs.setdefault('add_help', False)
            kwargs = self.get_primary_init_kwargs(**kwargs)
            parser = argparse.ArgumentParser(prog=self.prog_name, **kwargs)
            parser._name = self.primary_name
            # parser.add_argument('--foo', action='store_true', help='foo help')
            parser.set_defaults(func=self.default_caller)
            self.parser = parser

        return self.parser

    def get_primary_init_kwargs(self, **kwargs):
        nkw = self.primary_init_kwargs.copy()
        nkw.update(kwargs)
        return nkw

    def get_parser(self, name=None):
        if name in [None, "primary", self.primary_name]:
            return self.get_primary_parser()
        try:
            return self.positions[name]
        except KeyError as err:
            s = f'No key "{name}" - available: {self.positions.keys()}'
            raise NoPosition(s)

    def parse_args(self, args=None):
        aa, uu = self.get_parser().parse_known_args(args)
        print(uu)
        aa.unknown_args = uu
        self.args = args_unit = aa

        return args_unit

    def get_parsed_args(self, args=None):

        if self.args is None:
            return self.parse_args(args)
        return self.args

    def run_hook(self, *args, **kwargs):
        arg_unit = self.get_parsed_args()
        func = arg_unit.func
        return func(arg_unit, *args, **kwargs)



from django import template
from django.template.base import token_kwargs
from loguru import logger

from .base import SlotList, parse_until
from .slot_node import DefineSlotNode, SlotNode

log = logger.debug
warn = logger.warning


def do_wrap(parser, token):
    """Wrap the contents given to the nodem with another template given through
    the tag token:

        {% wrap "stocks/wrap_form.html" with button_text="Next"  action="/fo/bar" %}
            <ul>
            {% for row in form %}
                <li>{{ row }}</li>
            {% endfor %}
            </ul>
        {% endwrap %}

    The template "wrap_forms.html" should refer to the `wrap.content` and any
    other variables from the `with` or parent contexts.

        <form method="post"
          enctype="multipart/form-data"
          {% if action %}action="{{action}}"{% endif %}>

            {{ wrap.content }}
            <input class='btn'
                type="submit"
                value="{% firstof button_text 'Confirm' %}">
            {% csrf_token %}
        </form>
    """
    log("   _def_ do wrap")

    slotinfo = SlotList()
    slotdefine = SlotList()

    token.slotinfo = slotinfo
    token.slotdefine = slotdefine

    nodelist, splits, extra = parse_until(parser, token, ("endwrap",))

    res = WrappedContentNode(nodelist, splits, **extra)
    # res.slotinfo = slotinfo
    # res.slotdefine = slotdefine

    return res


class WrappedContentNode(template.Node):
    slotinfo = None
    slotdefine = None

    def __init__(self, nodelist, tokens, *a, **kw):
        log("  _cre_ __init__ wrap")
        self.nodelist = nodelist
        self.tokens = tokens
        self.token_template_name = template.Variable(tokens[0])
        self.extra_context = kw

    def announce_wrapper(self, wrap_templ):
        # wrap_templ.compile_nodelist()
        nodes = wrap_templ.nodelist
        res = ()
        # Iterate through the {% slots.define %} within the wrap template
        # Each node is given _this_ parent instance, allowing the DefineSlotNode
        # to read this _parent_wrap_ when generating its content.
        for node in nodes:
            if isinstance(node, DefineSlotNode):
                log(f" ! Found {type(node).__name__} slot {node.token_slot_name}")
                node.apply_parent(self)
                res += (node,)

        return res

    def render(self, context):
        slots = self.token.slotinfo

        log(f"\n\n\n    _ren_ Wrap node rendering template with: {slots}")
        ## The name supplied to the wrap object. {% wrap "name.html" %}
        ### Note (bug(: This doesn't work with relative paths.
        template_name = self.token_template_name.resolve(context)

        ## The target {% wrap name %} template - containing slots and the _result_.
        wrap_templ = context.template.engine.get_template(template_name)
        define_slots = self.announce_wrapper(wrap_templ)

        if len(define_slots) == 0:
            log(" - rendering standard...")
            # No slot.define, generic render can proceed.
            return self.standard_render(context, wrap_templ)
        log(f"Rendering with wrap slots: {define_slots}")
        return self.slot_render(
            context, wrap_templ, define_slots, template_name=template_name
        )

    def slot_render(self, context, wrap_templ, define_slots, **extras):
        # slotdefine = self.token.slotdefine

        ## We set any recently discovered (through the announce), slot.define
        # within the template. This is done within the render, as the wrap
        # template has only just instansiated, and the DefineSlotNodes hadn't
        # a chance in the past to declare themselves to the wrap.
        self.token.slotdefine.set(define_slots)

        self.populate_lost_content_slot(context, self.nodelist)
        return self.standard_render(context, wrap_templ)

    def filter_not(self, nodes, types):
        # Discover any nodes within this wrap - of which is not a slot type.
        # Strip these, as they should be pushed into the default, lost
        # area.
        # If there are no slots defined, the content is rendered as standard
        lost_nodes = ()
        for item in nodes:
            if isinstance(item, types):
                continue
            # found lost node.
            lost_nodes += (item,)
        return lost_nodes

    def populate_lost_content_slot(self, context, nodes):
        ## slots named "lost" in the define list.

        slotdefine = self.token.slotdefine
        slot_types = (
            DefineSlotNode,
            SlotNode,
        )
        lost_nodes = self.filter_not(nodes, slot_types)

        log(f"Found {len(lost_nodes)} lost nodes (without a wrap slot)")

        slotinfo = self.token.slotinfo

        target_slotname = "lost"
        if len(slotinfo) == 0:
            ## If the input slotcount is 0, everything into default.
            # If the output slots is 1, then it can only be that one slot.
            #
            # however, as _no-slots_ in, should perhaps  _not_ go to
            # slot "other-content" by accident.
            #
            # So _default_ (if exists) for unslotted content,
            # then _lost_ if no _default_,
            # fail if no _lost_ or _default_
            # (the slot.define are named or no define slots exist.)
            target_slotname = "default"

        target_slotdefs = slotdefine.get_nodes(target_slotname, context)

        log(f"Found {len(target_slotdefs)} slot.define for: {target_slotname}")

        ## if the target was missing, then the default or lost is not available.
        # As such if we tested 'default', finally test lost. Else quit.
        if len(target_slotdefs) == 0 and target_slotname == "default":
            log(f"No target '{target_slotname}', using _lost_.")
            target_slotname = "lost"
            target_slotdefs = slotdefine.get_nodes(target_slotname, context)

        if len(target_slotdefs) == 0:
            # No lost slot. Change the definition to another;
            warn(f"No define slots for _lost_ or _default_ - this is truly lost data.")
            return

        return
        # BUG: The lost slots are stacked into this for every request.
        # Likely due to this _add_ during the render phase.
        # And the SlotList is _sticky_ to the token, not this (fresh) Node.

        # The lost_nodes is a tuple, we need a NodeList type for the
        # newly generated info slot; lost
        NL = type(nodes)
        # Iterate all `slot.define lost` - Generally only one would exist.
        for name, unit in target_slotdefs:
            # push the lost content into the slotinfo, using the same name
            # as the _unit_, therefore the _lost_ or _default_ `slot.define`
            # render exactly like a named slot.
            simulated_lost_slot = SlotNode(NL(lost_nodes), unit.tokens)
            ## Once applied, the next render sequence captures this
            #'slotinfo' stack similar to a named slot.

            # BUG: The lost slots are stacked into this for every request.
            # Likely due to this _add_ during the render phase.
            # And the SlotList is _sticky_ to the token, not this (fresh) Node.
            slotinfo.add(simulated_lost_slot)

    def standard_render(self, context, wrap_templ, **extras):
        """Render the wrapper template, using the finalised self.nodelist

        At this point for a slotted template, the slot information
        has been shuffled into the correct nodes, or the node within the nodelist
        (such as a SlotNode) has been populated with the correct assets.

        """
        ## Render this {% wrap %}
        content = self.nodelist.render(context)

        sub_ctx = {
            "wrap": {
                "content": content,
                **extras,
            },
            "content": content,
        }

        values = {}
        for key, val in self.extra_context.items():
            values[key] = val.resolve(context) if hasattr(val, "resolve") else val

        with context.push(**sub_ctx, **values):
            return wrap_templ.render(context)



# Generated by Django 4.2.11 on 2024-05-02 01:35

import uuid

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("account", "0007_accountemail_uuid_token"),
    ]

    operations = [
        migrations.CreateModel(
            name="EmailInvite",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "email_address",
                    models.EmailField(
                        blank=True,
                        help_text="The target user email address",
                        max_length=254,
                        null=True,
                    ),
                ),
                (
                    "uuid_token",
                    models.CharField(
                        blank=True, default=uuid.uuid4, max_length=255, null=True
                    ),
                ),
                ("accepted", models.DateTimeField(blank=True, null=True)),
                ("expires", models.DateTimeField(blank=True, null=True)),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
    ]



"""
Test trim.signals module.

Simple tests for signal handlers and string formatters.
"""

import unittest
from unittest.mock import patch

from django.db.models import Model

from trim.signals import model_pre_init, repr_printer, str_printer


class StrPrinterTest(unittest.TestCase):
    """Test str_printer function."""

    def test_with_trim_string(self):
        # Setup
        class Obj:
            _trim_string = "My String"

        # Execute & Assert
        self.assertEqual(str_printer(Obj()), "My String")

    def test_with_get_method(self):
        # Setup
        class Obj:
            def get_trim_string(self):
                return "From Method"

        # Execute & Assert
        self.assertEqual(str_printer(Obj()), "From Method")

    def test_with_trim_props_single(self):
        # Setup
        class Obj:
            _trim_props = "name"
            name = "Alice"

        # Execute & Assert
        self.assertEqual(str_printer(Obj()), "Alice")

    def test_with_trim_props_tuple(self):
        # Setup
        class Obj:
            _trim_props = ("name", "age")
            name = "Bob"
            age = 25

        # Execute
        result = str_printer(Obj())

        # Assert
        self.assertIn("name=", result)
        self.assertIn("Bob", result)

    def test_with_alts(self):
        # Setup
        class Obj:
            _custom = "Custom"

        # Execute & Assert
        self.assertEqual(str_printer(Obj(), alts=("_custom",)), "Custom")

    def test_format_interpolation(self):
        # Setup
        class Obj:
            _trim_string = "User {self.name}"
            name = "Charlie"

        # Execute & Assert
        self.assertEqual(str_printer(Obj()), "User Charlie")


class ReprPrinterTest(unittest.TestCase):
    """Test repr_printer function."""

    def test_basic_format(self):
        # Setup
        class Product:
            pk = 123
            _trim_string = "Widget"

        # Execute & Assert
        self.assertEqual(repr_printer(Product()), "<Product(123) 'Widget'>")

    def test_with_trim_string_repr_alt(self):
        # Setup
        class Item:
            pk = 99
            _trim_string_repr = "Repr Format"
            _trim_string = "Normal"

        # Execute
        result = repr_printer(Item())

        # Assert
        self.assertIn("Repr Format", result)
        self.assertNotIn("Normal", result)


class ModelPreInitTest(unittest.TestCase):
    """Test model_pre_init signal handler."""

    @patch("trim.signals.hook_init_model_mixins")
    def test_sets_str_methods_with_trim_string(self, mock_hook):
        # Setup
        class FakeModel:
            _trim_string = "test"
            __str__ = Model.__str__
            __repr__ = Model.__repr__

        # Execute
        model_pre_init(FakeModel, args=(), kwargs={})

        # Assert
        self.assertEqual(FakeModel.__str__, str_printer)
        self.assertEqual(FakeModel.__repr__, repr_printer)

    @patch("trim.signals.hook_init_model_mixins")
    def test_sets_str_methods_with_trim_props(self, mock_hook):
        # Setup
        class FakeModel:
            _trim_props = ("field",)
            __str__ = Model.__str__
            __repr__ = Model.__repr__

        # Execute
        model_pre_init(FakeModel, args=(), kwargs={})

        # Assert
        self.assertEqual(FakeModel.__str__, str_printer)
        self.assertEqual(FakeModel.__repr__, repr_printer)

    @patch("trim.signals.hook_init_model_mixins")
    def test_doesnt_override_custom_methods(self, mock_hook):
        # Setup
        custom = lambda self: "custom"

        class FakeModel:
            _trim_string = "test"
            __str__ = custom
            __repr__ = custom

        # Execute
        model_pre_init(FakeModel, args=(), kwargs={})

        # Assert
        self.assertEqual(FakeModel.__str__, custom)
        self.assertEqual(FakeModel.__repr__, custom)


if __name__ == "__main__":
    unittest.main()



from django.apps import AppConfig
from django.conf import settings

from trim.apps import live_import

from . import context


class ThemingConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "trim.theming"

    def ready(self):
        context.build_default_themer()
        # Import /app/theming.py from all apps
        # print(' -- live importing theming')
        live_import("theming")



"""Wagtail 'blocks' in an easier form:

https://docs.wagtail.org/en/stable/reference/streamfield/blocks.html
"""

try:

    try:
        from wagtail.blocks import *
        from wagtail.fields import StreamField
    except ImportError as e:
        from wagtail.core.blocks import *
        from wagtail.core.fields import StreamField

    from wagtail.admin.panels import FieldPanel
    from wagtail.api import APIField
    from wagtail.documents.blocks import DocumentChooserBlock
    from wagtail.embeds.blocks import EmbedBlock
    from wagtail.images.blocks import ImageChooserBlock
    from wagtail.snippets.blocks import SnippetChooserBlock

    # StreamField
    PREPARED = True
except ImportError as e:
    PREPARED = False
    raise e

if PREPARED is True:
    streamfield = stream_field = StreamField


def field_panel(*a, **kw):
    return FieldPanel(*a, **kw)


def api_field(*a, **kw):
    return APIField(*a, **kw)


fieldpanel = field_panel
apifield = api_field


def chars(*a, **kw):
    return CharBlock(*a, **kw)


def integer(*a, **kw):
    return IntegerBlock(*a, **kw)


def decimal(*a, **kw):
    return DecimalBlock(*a, **kw)


def boolean(*a, **kw):
    return BooleanBlock(*a, **kw)


def datetime(*a, **kw):
    return DateTimeBlock(*a, **kw)


def text(*a, **kw):
    return TextBlock(*a, **kw)


def email(*a, **kw):
    return EmailBlock(*a, **kw)


def float_(*a, **kw):
    return FloatBlock(*a, **kw)


def regex(*a, **kw):
    return RegexBlock(*a, **kw)


def url(*a, **kw):
    return URLBlock(*a, **kw)


def date(*a, **kw):
    return DateBlock(*a, **kw)


def time(*a, **kw):
    return TimeBlock(*a, **kw)


def richtext(*a, **kw):
    return RichTextBlock(*a, **kw)


def rawhtml(*a, **kw):
    return RawHTMLBlock(*a, **kw)


def blockquote(*a, **kw):
    return BlockQuoteBlock(*a, **kw)


def choice(*a, **kw):
    return ChoiceBlock(*a, **kw)


def multiplechoice(*a, **kw):
    return MultipleChoiceBlock(*a, **kw)


def pagechooser(*a, **kw):
    return PageChooserBlock(*a, **kw)


def documentchooser(*a, **kw):
    return DocumentChooserBlock(*a, **kw)


def imagechooser(*a, **kw):
    return ImageChooserBlock(*a, **kw)


def snippetchooser(*a, **kw):
    return SnippetChooserBlock(*a, **kw)


def embed(*a, **kw):
    return EmbedBlock(*a, **kw)


def static(*a, **kw):
    return StaticBlock(*a, **kw)


def struct(*a, **kw):
    return StructBlock(*a, **kw)


def list_(*a, **kw):
    return ListBlock(*a, **kw)


def stream(*a, **kw):
    return StreamBlock(*a, **kw)


str = chars
float = float_
list = list_
dt = date_time = datetime
raw_html = rawhtml
rich_text = richtext
block_quote = blockquote
multiple_choice = multiplechoice
page_chooser = pagechooser
document_chooser = documentchooser
image_chooser = imagechooser
snippet_chooser = snippetchooser
stream_block = stream



from django import template
from django.template import Context
from django.template.base import token_kwargs
from loguru import logger

log = logger.debug


def inject_node(parser):
    T = parser.tokens[0].__class__
    tr = "{{ slotplot }}"
    parser.prepend_token(T(parser.tokens[0].token_type, tr))


def parse_until(parser, token, ends, delete_first=True):
    """`parser` is the iterative unit processing the nodes
    The `token` represents the current block, such as '<block token "wrap">'
    `ends` are a list of end-tokens to stop the parser

    """
    nodelist = parser.parse(ends)
    if delete_first:
        parser.delete_first_token()
    # inject_node(parser)
    splits = token.split_contents()[1:]
    extra = {}
    for i, word in enumerate(splits):
        if word == "with":
            extra = token_kwargs(splits[i + 1 :], parser)
    return nodelist, splits, extra


class SlotList(object):
    def __init__(self, *a, **kw):
        log("  _def_ new SlotList", id(self))
        self.units = ()

    def add(self, unit):
        # log(' -- SlotList.add', unit)
        self.units += (unit,)

    def set(self, iterable):
        self.units = tuple(iterable)

    def apply_lost(self, nodelist, context=None):
        """Apply the given nodelist to the 'lost' node, for
        the incoming render request.
        """
        log(f"{len(nodelist)} items")
        lost_slots = self.get_nodes("lost", context)
        for name, unit in lost_slots:
            log(f"Slot found: {unit}")

    def get_nodes(self, first_token, context=None):
        """Return the target node given the name, this
        could be None, "default" or _named_
        """
        res = ()

        for unit in self.units:
            names = unit.get_slot_names(context)
            if first_token in names:
                log(f"Found match! {first_token}")
                res += (
                    (
                        first_token,
                        unit,
                    ),
                )

        return res  # default_value # content = slotlist.

    def __len__(self):
        return len(self.units)

    def __repr__(self):
        return f"<{self.__class__.__name__} {len(self.units)} slots>"



from django import forms

from trim.forms import fields


class ListForm(forms.Form):

    count = fields.integer(required=False)
    order_by = fields.chars(required=False)
    direction = fields.chars(required=False)



"""Django form fields as a drop-in for existing fields.
Mix and match with existing components to transition through trim bits

    from trim.forms import fields

    class ProductQuestionForm(forms.Form):
        product_id = forms.CharField(widget=fields.widgets.hidden())
        message = fields.text(attrs={
                'placeholder': 'Does this include a spare flux-capacitor?'
            })

"""

from django import forms

from . import widgets


def boolean(*a, **kw):
    """A standard `forms.BooleanField` field."""
    return forms.BooleanField(*a, **kw)


def boolean_false(*a, **kw):
    """A standard `forms.BooleanField` field with the `required` as `False`
    for a boolean with a default False value."""
    kw.setdefault("required", False)
    return boolean(*a, **kw)


def boolean_true(*a, **kw):
    """A standard `forms.BooleanField` field with the `required` as `False`
    for a boolean with a default False value."""
    kw.setdefault("initial", True)
    return boolean(*a, **kw)


def chars(*a, **kw):
    """A standard `forms.CharField` field."""
    return forms.CharField(*a, **kw)


def hidden_chars(*a, **kw):
    kw.setdefault("widget", widgets.hidden())
    return chars(*a, **kw)


def password(*a, **kw):
    """A standard `forms.CharField` field. with a password input widget"""
    widget = widgets.password()
    kw.setdefault("widget", widget)
    return char(*a, **kw)


def text(*a, attrs=None, rows=3, cols=30, **kw):
    """A standard `forms.CharField` field. with a TextArea input widget"""
    attrs = attrs or {"rows": rows, "cols": cols}
    widget = widgets.textarea(attrs=attrs)
    kw.setdefault("widget", widget)
    return char(*a, **kw)


def choice(*a, **kw):
    """A standard `forms.ChoiceField` field."""
    return forms.ChoiceField(*a, **kw)


def date(*a, **kw):
    """A standard `forms.DateField` field."""
    return forms.DateField(*a, **kw)


def datetime(*a, **kw):
    """A standard `forms.DateTimeField` field."""
    return forms.DateTimeField(*a, **kw)


def decimal(*a, **kw):
    """A standard `forms.DecimalField` field."""
    return forms.DecimalField(*a, **kw)


def duration(*a, **kw):
    """A standard `forms.DurationField` field."""
    return forms.DurationField(*a, **kw)


# def email(*a, **kw):
#     """A standard `forms.EmailField` field."""
#     attrs = _pop('attrs', kw)
#     res = forms.EmailField(*a,**kw)
#     _push(attrs, res.widget.attrs.update)
#     return res


def email(*a, **kw):
    """A standard `forms.EmailField` field."""
    widget_update = lambda field, data: field.widget.attrs.update(data)
    res = _pop_push("attrs", widget_update, forms.EmailField, *a, **kw)
    return res


def _pop_push(field, pushed, callable_unit, *a, **kw):
    """Pop and push back the field onto the callable unit
    _pop_push('attrs', res.widget.attrs.update,
                forms.EmailField, *a,**kw)
    """
    attrs = _pop(field, kw)
    res = callable_unit(*a, **kw)
    _push(attrs, pushed, res)
    return res


def _pop(key, kw):
    return kw.pop(key) if key in kw else None


def _push(data, lam, *a):
    return lam(*a, data) if data is not None else None


def file(*a, **kw):
    """A standard `forms.FileField` field."""
    return forms.FileField(*a, **kw)


def files(*a, **kw):
    """A _muliple_ Standard form file field, with a clearable input."""
    return file(widget=widgets.clearable_file_input(attrs={"multiple": True}))


def file_path(*a, **kw):
    """A standard `forms.FilePathField` field."""
    return forms.FilePathField(*a, **kw)


def float(*a, **kw):
    """A standard `forms.FloatField` field."""
    return forms.FloatField(*a, **kw)


def generic_ip_address(*a, **kw):
    """A standard `forms.GenericIPAddressField` field."""
    return forms.GenericIPAddressField(*a, **kw)


def image(*a, **kw):
    """A standard `forms.ImageField` field."""
    return forms.ImageField(*a, **kw)


def integer(*a, **kw):
    """A standard `forms.IntegerField` field."""
    return forms.IntegerField(*a, **kw)


def json(*a, **kw):
    """A standard `forms.JSONField` field."""
    return forms.JSONField(*a, **kw)


def multiple_choice(*a, **kw):
    """A standard `forms.MultipleChoiceField` field."""
    return forms.MultipleChoiceField(*a, **kw)


def null_boolean(*a, **kw):
    """A standard `forms.NullBooleanField` field."""
    return forms.NullBooleanField(*a, **kw)


def regex(*a, **kw):
    """A standard `forms.RegexField` field."""
    return forms.RegexField(*a, **kw)


def slug(*a, **kw):
    return forms.SlugField(*a, **kw)


def time(*a, **kw):
    """A standard `forms.TimeField` field."""
    return forms.TimeField(*a, **kw)


def typed_choice(*a, **kw):
    """A standard `forms.TypedChoiceField` field."""
    return forms.TypedChoiceField(*a, **kw)


def typed_multiple_choice(*a, **kw):
    """A standard `forms.TypedMultipleChoiceField` field.

    https://docs.djangoproject.com/en/5.0/ref/forms/fields/#typedmultiplechoicefield
    """
    return forms.TypedMultipleChoiceField(*a, **kw)


def url(*a, **kw):
    """A standard `forms.URLField` field.

    https://docs.djangoproject.com/en/5.0/ref/forms/fields/#urlfield
    """
    return forms.URLField(*a, **kw)


def uuid(*a, **kw):
    """A standard `forms.UUIDField` field.

    https://docs.djangoproject.com/en/5.0/ref/forms/fields/#uuidfield
    """
    return forms.UUIDField(*a, **kw)


def combo(*a, **kw):
    """A standard `forms.ComboField` field.

        ComboField(fields=[CharField(max_length=20), EmailField()])

    https://docs.djangoproject.com/en/5.0/ref/forms/fields/#combofield
    """
    return forms.ComboField(*a, **kw)


def multi_value(*a, **kw):
    """A standard `forms.MultiValueField` field.

    https://docs.djangoproject.com/en/5.0/ref/forms/fields/#multivaluefield
    """
    return forms.MultiValueField(*a, **kw)


def split_datetime(*a, **kw):
    """A standard `forms.SplitDateTimeField` field.

    https://docs.djangoproject.com/en/5.0/ref/forms/fields/#splitdatetimefield
    """
    return forms.SplitDateTimeField(*a, **kw)


def modelchoice(*a, **kw):
    """A standard `forms.ModelChoiceField` field.

    https://docs.djangoproject.com/en/5.0/ref/forms/fields/#modelchoicefield
    """
    return forms.ModelChoiceField(*a, **kw)


def hidden(field=None, **kwargs):
    """Wrap a callable or field instance with `hidden()` to automatically
    apply a _hidden_ widget.

    If the given field is callable, the function is called before applying
    the `widgets.hidden` to the field instance.

        class StockNotifyForm(forms.Form):
            product_id = fields.hidden(fields.chars)
            email = fields.email()

    """
    if field is None:
        return hidden_chars(**kwargs)
    if callable(field):
        field = field()
    field.widget = widgets.hidden(**kwargs)
    return field


img = image
int = integer
pwd = password
bool = boolean
false_bool = bool_false = boolean_false
true_bool = bool_true = boolean_true
str = char = chars
chars_hidden = chars
hide = hidden
textarea = text
file_multi = multi_file = files



from django.utils.functional import cached_property
from django.views.generic import ListView

from ..forms import list as forms


class OrderPaginatedListView(ListView):
    """A ListView with prepared ordering and pagination."""

    default_orderby = "default_orderby_field"
    default_selected_orderby = "name"
    default_direction = "asc"
    ordering = f"{default_orderby}"

    min_paginate_by = 10
    max_paginate_by = 200
    paginate_by = 50

    order_by_field = "order_by"
    paginate_by_field = "count"
    direction_field = "direction"

    ordering_fields = (
        # ext. val, ext. label, int. key
        ("name", ("Name", "short_description")),
        (
            "cost",
            (
                "Price",
                "cached_current_price",
            ),
        ),
        ("product", ("Product ID", "product_id")),
    )

    direction_fields = (
        (
            "asc",
            "",
        ),
        (
            "desc",
            "-",
        ),
    )

    def get_ordering_fields(self, expand=False):
        r = ()
        for k, v in self.ordering_fields:
            if (not expand) and isinstance(v, (tuple, list)):
                v = v[0]
            r += (
                (
                    k,
                    v,
                ),
            )
        return r

    def get_form(self):
        return forms.ListForm(self.request.GET)

    @cached_property
    def filter_data(self):
        return self.get_filter_data()

    def get_filter_data(self):
        f = self.get_form()
        if f.is_valid():
            d = f.cleaned_data
            d = self.format_internal(d)
            return d

        return {}

    def get_orderby_field(self, d):
        field = d.get("order_by", self.default_orderby)
        row = dict(self.get_ordering_fields(expand=True)).get(field)
        if row is None:
            # bad field.
            return self.default_orderby
        if isinstance(row, tuple):
            return row[1]
        return row
        # return field

    def get_selected_orderby(self):
        return (
            self.filter_data.get(self.order_by_field, None)
            or self.default_selected_orderby
        )

    def get_selected_direction(self):
        return (
            self.filter_data.get(self.direction_field, None) or self.default_direction
        )

    def format_internal(self, data):
        """Convert the form dict into something internal."""
        field = self.get_orderby_field(data)
        direction = data.get(self.direction_field, None)
        direction = dict(self.direction_fields).get(direction, None) or ""
        # direction = '' if direction else '-'

        ds = f"{direction}{field}"
        data["ordering"] = ds
        print("Order", ds)
        data.setdefault("paginate_by", self.paginate_by)
        user_pagination = self.clamp_paginate_by(
            data.get(self.paginate_by_field, None) or None
        )
        if user_pagination is not None:
            data["paginate_by"] = user_pagination
        return data

    def get_ordering(self):
        v = self.filter_data.get("ordering", "")
        if v is None or len(v) == 0:
            return self.ordering
        return v

    def get_paginate_by(self, queryset=None):
        v = self.paginate_by
        r = self.clamp_paginate_by(v)
        return self.filter_data.get("paginate_by", r)

    def clamp_paginate_by(self, v):
        return min(max(self.min_paginate_by, v or 0), self.max_paginate_by)



from django.conf import settings

## Baked 'settings' for the theming unit. These vars stay bound outside
# the options of the template usage. In some cases (such as `version`) the
# option may be overridden - but generally these should be static during
# runtime.
THEMING_OPTIONS = {
    "root": "theme",
    "version": "",
}

THEMING_MAP_DEFAULTS = {
    "form": "form.html",
    "page": "page.html",
    "list": "list.html",
    "detail": "detail.html",
}

## The name -> file map exposed through the {% theme %} and {{ theme }}
# template tags. This is mapped into the _forced_ theming options, ensuring
# the templates may call _into_ one directory. By default this is the given
# "theme"
THEMING_MAP = getattr(
    settings,
    "THEMING_MAP",
    {
        # 'detail.double': 'detail-double-header.html',
        # 'detail.default': 'detail.html',
    },
)

from collections import ChainMap

LIVE_MAPS = {"ready_map": ChainMap(THEMING_MAP, THEMING_MAP_DEFAULTS)}


def set_ready_map(data):
    LIVE_MAPS["ready_map"] = data


def get_theme_map():
    return LIVE_MAPS["ready_map"]


def name_redirect(origin, target):
    THEMING_MAP.setdefault(origin, get_theme_map()[target])


def name_default_redirect(
    origin,
    target,
    target_template,
    default_template=None,
    default_word="default",
    sep=".",
):
    """Build a redirection of a default template name, to a special template
    name, "detail" -> "detail.other", with the _other_ as the new primary.

        name_default_redirect('detail', 'other', 'detail-other.html',)

    Producing a THEME entry:

        'detail.other' = 'detail-other.html'
        'detail.default' = 'detail.html'

    Functionally this is similar to:

        THEMING_MAP = {
            'detail.double': 'detail-double-header.html',
            'detail.default': 'detail.html',
        }
        THEMING_MAP.setdefault('detail', THEMING_MAP['detail.double'])

    """
    # Create the 'detail.default' entry
    default_key = sep.join([origin, default_word])
    default_template = default_template or get_theme_map().get(origin)
    THEMING_MAP.setdefault(default_key, default_template)

    # Create the _new_ entry
    new_key = sep.join([origin, target])
    THEMING_MAP.setdefault(new_key, target_template)

    # Add the redirect "detail" => "detail.double"
    return name_redirect(origin, new_key)


## Install a _default redirect_, allowing the import of the "detail"
# and it invisibly resolves "detail.double" with the addition "details.default"
# created automatically.
# name_default_redirect('detail', 'double', 'detail-double-header.html',)

## for base variants its nice to set
# the default to an inner key.
# The inner key may be prone to change, but the outer key (detail)
# should stay stable to the template.
# THEMING_MAP.setdefault('detail', THEMING_MAP['detail.double'])

UNDEF = {}


def theme_option(key, var=UNDEF, default=None):
    if var in (None, UNDEF):
        OPTS = getattr(settings, "THEMING_OPTIONS", None) or THEMING_OPTIONS
        return OPTS.get(key, default)
    return var


def install(*maps, **kw):
    """Install a theme pack from the user `theming.py` loadout."""
    ready_map = get_theme_map()
    scope = ready_map.new_child()
    for md in maps:
        scope.update(md)
    set_ready_map(scope)
    # print('trim.theming.config.install', ready_map)
    return scope



"""
Register your models in the admin -

    from django.contrib import admin
    from trim import admin as trims

    from . import models

    trims.register_models(models)

"""

from django.contrib import admin

# Register your models here.
from .models import cache_known, grab_models


def register_models(models, ignore=None):
    """Register a list of models within the standard admin site register,
    calling upon trim.models.grab_models() to detect the inbound models.

    Provide an ignore list to omit models from the register; in such cases as
    model pollution from app imports, or modeladmin customisation.

    string method:

        trims.register_models(models, ignore=['GenericProduct'])

    models:

        trims.register_models(models, ignore=[models.GenericProduct])

    With model register detection:

        trims.register_models(models, ignore=__name__)

    Use `ignore=__name__` in conjunction with the `@trim.admin.register`
    """
    return admin.site.register(grab_models(models, ignore=ignore))


def register(*a, **kw):
    """Perform a standard django `admin.register(Model)` with the model being
    _cached_ for later assessment during the grab models ignore filters.

    Replace `admin.register` with `trims.register`:

        from django.contrib import admin
        from trim import admin as trims
        from . import models

        # Standard
        @admin.register(models.GenericProduct):
        class GenericProductAdmin(admin.ModelAdmin):
            pass


    Exactly the same through trims:

        from django.contrib import admin
        from trim import admin as trims
        from . import models

        @trims.register(models.GenericProduct)
        class GenericProductAdmin(admin.ModelAdmin):
            pass

    Utilising `trims.register` captures the used model, therefore when
    registering generic model views, the `ignore` method may identify the module
    names to exclude:

    The standard django method:

        from . import models
        admin.site.register(models.GenericProduct,
                            models.OtherModel,
                            models.AnotherModel, ...)


    trim method allows importing of _all_ models:

        from . import models
        trims.register_models(models)

    Ignore previously registered models:

        trims.register_models(models,
                ignore=['GenericProduct', models.Anothermodel])


    or better-yet, use the model register. The `ignore` attribute identifies
    the _current_ module e.g. `products.admin`.

        # within admin.py
        trims.register_models(models, ignore=__name__)

        ## Synonymous to
        trims.register_models(models, ignore='products.admin')

    Any model previously utilised by `@trims.register` for the products module
    is automatically exlcuded.
    """
    cache_known(*a)
    return admin.register(*a, **kw)



from django.core.management.base import BaseCommand, CommandError

from trim.models import django

# from polls.models import Question as Poll


class Command(BaseCommand):
    help = "Closes the specified poll for voting"

    # def add_arguments(self, parser):
    # parser.add_argument('poll_ids', nargs='+', type=int)

    # def handle(self, *args, **options):
    #     for poll_id in options['poll_ids']:
    #         try:
    #             poll = Poll.objects.get(pk=poll_id)
    #         except Poll.DoesNotExist:
    #             raise CommandError('Poll "%s" does not exist' % poll_id)

    #         poll.opened = False
    #         poll.save()

    #         self.stdout.write(self.style.SUCCESS('Successfully closed poll "%s"' % poll_id))

    def handle(self, *args, **options):
        self.out("Handle gen doc")

    def out(self, *a):
        self.stdout.write(self.style.SUCCESS(" ".join(map(str, a))))



from django.contrib.auth import logout, update_session_auth_hash
from django.contrib.auth.forms import PasswordChangeForm
from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.auth.views import LoginView, LogoutView
from django.shortcuts import render
from django.urls import reverse_lazy

from trim import views as shorts

from .. import models


class PasswordChangeView(LoginRequiredMixin, shorts.FormView):
    model = shorts.get_user_model()
    form_class = PasswordChangeForm
    success_url = reverse_lazy("account:profile")

    template_name = "account/user_password_change.html"

    def form_valid(self, form):
        form.save()
        update_session_auth_hash(self.request, form.user)
        return super().form_valid(form)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["user"] = self.request.user
        if self.request.method == "POST":
            kwargs["data"] = self.request.POST
        return kwargs


class ProfilePasswordUpdateView(PasswordChangeView):
    # model = CustomUser
    template_name = "account/password_update_form.html"


def logout_view(request):
    logout(request)


class ProfileLogout(LogoutView):
    template_name = "account/logged-out.html"


class ProfileLogin(LoginView):
    template_name = "account/login.html"


class ProfileInactiveAccount(shorts.TemplateView):
    template_name = "account/inactive.html"


from django.contrib.auth import views as auth_views


class PasswordResetView(auth_views.PasswordResetView):
    template_name = "account/forgotpasswordrecord_form.html"

    def form_valid(self, form):
        email = form.cleaned_data["email"]
        # If the email address is a registered user, send the email.
        U = shorts.get_user_model()
        email_field_name = U.get_email_field_name()
        # exists = U.objects.filter(email__iexact=email).exists()
        active_users = U._default_manager.filter(
            **{
                "%s__iexact" % email_field_name: email,
                "is_active": True,
            }
        )

        models.ForgotPasswordRecord.objects.create(email_address=email)

        for u in active_users:
            hup = u.has_usable_password()
            s = f"PasswordResetView::Requested email exists? has_usable_password: {hup}"
            print(s, u, email)

        return super().form_valid(form)


# class ProfileForgotPasswordView(shorts.CreateView):
#     model = models.ForgotPasswordRecord
#     fields = ('email_address',)
#     success_url = reverse_lazy('account:forgot_success')

#     def form_valid(self, form):
#         email = form.cleaned_data['email_address']
#         # If the email address is a registered user, send the email.
#         U = shorts.get_user_model()
#         exists = U.objects.filter(email__iexact=email).exists()
#         print('Requested email exists?', email, exists)
#         if exists:
#             self.send_reset_email(email)
#         return super().form_valid(form)

#     def send_reset_email(self, email_address):
#         # password reset here.
#         print('send email!')


class ProfileForgotPasswordSuccessView(shorts.TemplateView):
    template_name = "account/password-reset-success-view.html"



from django.apps import AppConfig
from django.db.models.signals import pre_init

from .models.auto import hook_waiting_model_mixins


class ShortConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "trim"

    def ready(self):
        # Implicitly connect a signal handlers decorated with @receiver.
        from . import signals

        hook_waiting_model_mixins()
        # Explicitly connect a signal handler.
        pre_init.connect(signals.model_pre_init)


import importlib

from django import apps
from django.conf import settings


def live_import(module_name):
    """Perform late imports of the target module name within each app
    to ensure any _lazy_ are collected.

    For example, import *.coolapp module for every installed app:

        class ProjectConfig(AppConfig):
            default_auto_field = 'django.db.models.BigAutoField'
            name = 'project'

            def ready(self):
                live_import('coolapp')

    This allows the import of modules for an installed app without mandatory imports
    within the target app.
    """

    _apps = apps.registry.apps
    res = ()
    for conf in _apps.get_app_configs():
        name = conf.label
        # print(' -- App', name)
        # name == 'admin'
        # __name__ == 'django.contrib.admin'
        package_name = conf.module.__name__
        mod = silent_import_package_module(package_name, module_name)
        if mod is not None:
            res += (mod,)
    ## Import the root app (it's not one of the installed apps.)
    a = settings.ROOT_URLCONF.split(".")[0]
    silent_import_package_module(a, module_name)

    return res


def silent_import_package_module(package_name, module_name):
    n = f"{package_name}.{module_name}"
    try:
        # print('  Looking for', n)
        vv = importlib.import_module(n)
        # print(' -! Imported', vv)
        return vv
    except ModuleNotFoundError as err:
        if err.name != n:
            # module name mismatch, the failed import is not the
            # one we requested.
            raise err



from django import forms

from trim.forms import fields


class AllFieldsForm(forms.Form):
    boolean = fields.boolean()
    boolean_false = fields.boolean_false()
    boolean_true = fields.boolean_true()
    chars = fields.chars()
    hidden_chars = fields.hidden_chars()
    password = fields.password()
    text = fields.text()
    choice = fields.choice()
    date = fields.date()
    datetime = fields.datetime()
    decimal = fields.decimal()
    duration = fields.duration()
    email = fields.email()
    file = fields.file()
    files = fields.files()
    file_path = fields.file_path()
    float = fields.float()
    generic_ip_address = fields.generic_ip_address()
    image = fields.image()
    integer = fields.integer()
    json = fields.json()
    multiple_choice = fields.multiple_choice()
    null_boolean = fields.null_boolean()
    regex = fields.regex()
    slug = fields.slug()
    time = fields.time()
    typed_choice = fields.typed_choice()
    typed_multiple_choice = fields.typed_multiple_choice()
    url = fields.url()
    uuid = fields.uuid()
    combo = fields.combo(fields=[fields.chars(), fields.email()])
    multi_value = fields.multi_value()
    split_datetime = fields.split_datetime()
    modelchoice = fields.modelchoice()
    hidden = fields.hidden()



from django.http import Http404
from django.shortcuts import redirect
from django.urls import reverse


class Missing404RedirectUrl(Exception):
    pass


from django.views.generic import TemplateView


class Custom404TemplateView(TemplateView):
    status_code = 404
    template_name = "trim/errors/404.html"

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs["status_code"] = self.status_code
        return kwargs


class Custom404(object):
    """A View mixin to capture disptach 404 and respond with a redirect
    or alternative view

        from trim import views

        class ProductBrandDetailView(views.errors.Custom404, views.DetailView):
            custom_404_redirect_url = 'appname:view'
            custom_404_redirect = True if redirect_url else False
            custom_404_template_name = 'appname/errors/special.html'

            model = models.ProductBrand
            ...

    """

    custom_404_redirect_url = None
    custom_404_redirect = None
    # custom_404_template_name

    def dispatch(self, request, *args, **kwargs):
        try:
            return super().dispatch(request, *args, **kwargs)
        except Http404:
            return self.custom_404(request, *args, **kwargs)

    def custom_404(self, request, *args, **kwargs):
        """Retutn a 404 response either as a redirect, or inline view replacement.
        If the redirect url exists the user is redirected, if missing, the
        function returns Custom404TemplateView as_view
        """
        if (self.custom_404_redirect_url is None) and (
            self.custom_404_redirect is False
        ):
            raise Missing404RedirectUrl("custom_404_redirect_url not supplied.")
        custom_404_redirect = self.custom_404_redirect
        if custom_404_redirect is None:
            if self.custom_404_redirect_url is not None:
                custom_404_redirect = True

        if custom_404_redirect is True:
            resolved = self.get_custom_404_url(request, *args, **kwargs)
            return self.custom_404_redirect_response(resolved)
        return self.custom_404_view(request, *args, **kwargs)

    def custom_404_view(self, request, *args, **kwargs):
        props = {}
        n = getattr(self, "custom_404_template_name", None)
        if n is not None:
            props["template_name"] = self.custom_404_template_name

        ct = Custom404TemplateView.as_view(**props)
        v = ct(request, *args, **kwargs)
        return v

    def get_custom_404_url(self, request, *args, **kwargs):
        if self.custom_404_redirect_url is None:
            raise Missing404RedirectUrl("custom_404_redirect_url not supplied.")
        return reverse(self.custom_404_redirect_url, args=args, kwargs=kwargs)

    def custom_404_redirect_response(self, url):
        return redirect(url)
        # redirect(reverse('products:filter', args=args, kwargs=kwargs))



"""
{% timedelta cart.updated cart.created format="" %}
{{ cart.created|timesince:cart.updated }}
{% human_timedelta cart.updated cart.created format="" %}
"""

from pydoc import locate

from django import template

register = template.Library()


@register.simple_tag(takes_context=False)
def functional(name, *args, **kwargs):
    """Dynamically call any Python function by its fully qualified name.

    This powerful template tag allows you to execute any accessible Python
    function directly from your templates by providing its module path and name.
    Uses `pydoc.locate` to find and execute the function with provided arguments.

    Args:
        name (str): Fully qualified function name (e.g., 'os.path.join', 'math.sqrt')
        *args: Positional arguments to pass to the function
        **kwargs: Keyword arguments to pass to the function

    Returns:
        Result of the function call, or empty string if function not found,
        or string representation if the located object is not callable

    Examples:
        {% load functional %}

        {# Call math functions #}
        {% functional 'math.sqrt' 16 %}
        {# Returns: 4.0 #}

        {# Call string methods #}
        {% functional 'str.upper' 'hello world' %}
        {# Returns: 'HELLO WORLD' #}

        {# Call custom module functions #}
        {% functional 'myapp.utils.format_price' product.price currency='USD' %}

        {# Call datetime functions #}
        {% functional 'datetime.datetime.now' %}

        {# Access module constants #}
        {% functional 'math.pi' %}
        {# Returns: 3.141592653589793 #}

    Security Warning:
        Use this tag with extreme caution in production environments.
        It can execute arbitrary code and poses security risks if the
        function name comes from untrusted sources. Only use with
        controlled, whitelisted function names in production.

    Note:
        If the located object is not callable (e.g., a constant or variable),
        it will be converted to a string and returned.
    """
    # Note: This is a simple tag, not a filter, so it does not take
    # the context or return a string to be rendered in the template.
    # It directly returns the result of the callable.
    _callable = locate(name)
    if not _callable:
        return ""

    if not callable(_callable):
        return str(_callable)

    return _callable(*args, **kwargs)



"""
Testing utilities for django-trim.

Provides helpers to validate view configurations, mixin order, and other
common patterns when using trim shortcuts.

Example usage in your tests:

    from trim import tests as trim_tests
    from myapp import views

    class ViewConfigTest(TestCase):
        def test_mixin_order(self):
            trim_tests.assert_mixin_order(views.MySecureView)

        def test_all_views(self):
            trim_tests.assert_views_mixin_order(views)
"""

from django.contrib.auth.mixins import (
    AccessMixin,
    LoginRequiredMixin,
    PermissionRequiredMixin,
    UserPassesTestMixin,
)
from django.test import TestCase
from django.views.generic.base import View

# Authentication/permission mixins that should come before view classes
AUTH_MIXINS = (
    AccessMixin,
    LoginRequiredMixin,
    PermissionRequiredMixin,
    UserPassesTestMixin,
)


class MixinOrderError(AssertionError):
    """Raised when mixins are in incorrect order."""

    pass


def get_mro_position(klass, target_class):
    """
    Get the position of a class in the Method Resolution Order.

    Args:
        klass: The class to inspect
        target_class: The class to find in the MRO

    Returns:
        int: Position in MRO, or -1 if not found
    """
    try:
        return klass.__mro__.index(target_class)
    except ValueError:
        return -1


def find_auth_mixins(klass):
    """
    Find all authentication/permission mixins in a view class.

    Args:
        klass: The view class to inspect

    Returns:
        list: List of tuples (mixin_class, position_in_mro)
    """
    found = []
    for mixin in AUTH_MIXINS:
        pos = get_mro_position(klass, mixin)
        if pos >= 0:
            found.append((mixin, pos))
    return found


def find_view_base(klass):
    """
    Find the base Django view class in the MRO.

    Args:
        klass: The view class to inspect

    Returns:
        tuple: (view_class, position) or (None, -1) if not found
    """
    for i, cls in enumerate(klass.__mro__):
        # Check if it's a Django generic view (but not View itself at position 0)
        if i > 0 and issubclass(cls, View) and cls is not View:
            # Found a concrete view class
            return (cls, i)
    return (None, -1)


def assert_mixin_order(view_class, raise_exception=True):
    """
    Assert that authentication/permission mixins appear before view classes
    in the Method Resolution Order.

    This ensures mixins can properly intercept and control access before
    the view logic executes.

    Args:
        view_class: The view class to validate
        raise_exception: If True, raises MixinOrderError on failure.
                        If False, returns (is_valid, error_message)

    Returns:
        tuple: (True, None) if valid, or (False, error_message) if raise_exception=False

    Raises:
        MixinOrderError: If mixins are in wrong order and raise_exception=True

    Example:
        >>> from trim.tests import assert_mixin_order
        >>> from trim import views
        >>>
        >>> class MyView(views.IsStaffMixin, views.UpdateView):
        ...     model = MyModel
        >>>
        >>> assert_mixin_order(MyView)  # Passes
        (True, None)

        >>> class BadView(views.UpdateView, views.IsStaffMixin):
        ...     model = MyModel
        >>>
        >>> assert_mixin_order(BadView)  # Raises MixinOrderError
    """
    if not issubclass(view_class, View):
        error_msg = f"{view_class.__name__} is not a Django view class"
        if raise_exception:
            raise TypeError(error_msg)
        return (False, error_msg)

    auth_mixins = find_auth_mixins(view_class)
    view_base, view_pos = find_view_base(view_class)

    if not auth_mixins:
        # No auth mixins found, nothing to validate
        return (True, None)

    if view_pos < 0:
        # No view base found (shouldn't happen, but defensive)
        return (True, None)

    # Check each auth mixin position
    problems = []
    for mixin_class, mixin_pos in auth_mixins:
        if mixin_pos > view_pos:
            problems.append(
                f"{mixin_class.__name__} (position {mixin_pos}) "
                f"appears AFTER {view_base.__name__} (position {view_pos})"
            )

    if problems:
        mro_display = " -> ".join([cls.__name__ for cls in view_class.__mro__[:6]])
        error_msg = (
            f"Incorrect mixin order in {view_class.__name__}:\n"
            f"  MRO: {mro_display}\n"
            f"  Problems:\n    - " + "\n    - ".join(problems) + "\n"
            f"  Fix: Place auth mixins BEFORE view classes:\n"
            f"    class {view_class.__name__}(AuthMixin, ViewClass):"
        )

        if raise_exception:
            raise MixinOrderError(error_msg)
        return (False, error_msg)

    return (True, None)


def assert_views_mixin_order(module, raise_exception=True):
    """
    Validate mixin order for all view classes in a module.

    Scans the module for Django view classes and validates each one.

    Args:
        module: Python module containing view classes
        raise_exception: If True, raises on first error. If False,
                        returns list of all errors found.

    Returns:
        list: Empty if all valid, or list of error messages if raise_exception=False

    Raises:
        MixinOrderError: If any view has incorrect order and raise_exception=True

    Example:
        >>> from trim.tests import assert_views_mixin_order
        >>> from myapp import views
        >>>
        >>> # In your test file:
        >>> class ViewTests(TestCase):
        ...     def test_all_views_mixin_order(self):
        ...         assert_views_mixin_order(views)
    """
    import inspect

    errors = []

    for name, obj in inspect.getmembers(module):
        # Check if it's a class and a Django view
        if inspect.isclass(obj) and issubclass(obj, View):
            # Skip the base View class itself
            if obj is View:
                continue

            is_valid, error = assert_mixin_order(obj, raise_exception=False)
            if not is_valid:
                if raise_exception:
                    raise MixinOrderError(error)
                errors.append(error)

    return errors


def get_view_permissions(view_class):
    """
    Extract permission requirements from a view class.

    Args:
        view_class: The view class to inspect

    Returns:
        dict: Dictionary with permission info:
            - 'permissions': list of required permissions
            - 'any': bool, whether ANY permission is sufficient (vs ALL)
            - 'login_required': bool
            - 'staff_required': bool

    Example:
        >>> info = get_view_permissions(MySecureView)
        >>> print(info['permissions'])
        ['myapp.change_model', 'myapp.delete_model']
    """
    info = {
        "permissions": [],
        "any": False,
        "login_required": False,
        "staff_required": False,
        "user_owned": False,
        "user_field": None,
    }

    # Check for PermissionRequiredMixin
    if hasattr(view_class, "permission_required"):
        perms = view_class.permission_required
        if isinstance(perms, str):
            info["permissions"] = [perms]
        else:
            info["permissions"] = list(perms)

        # Check if ANY permission is sufficient
        info["any"] = getattr(view_class, "permission_required_any", False)

    # Check for LoginRequiredMixin
    info["login_required"] = issubclass(view_class, LoginRequiredMixin)

    # Check for staff requirement (trim's IsStaffMixin uses UserPassesTestMixin)
    # We detect by checking for test_func that checks is_staff
    if hasattr(view_class, "test_func"):
        # This is a heuristic - can't perfectly detect without running
        info["staff_required"] = True

    # Check for UserOwnedMixin
    if hasattr(view_class, "user_field"):
        info["user_owned"] = True
        info["user_field"] = view_class.user_field

    return info


def assert_view_has_permission(view_class, permission):
    """
    Assert that a view requires a specific permission.

    Args:
        view_class: The view class to check
        permission: Permission string (e.g., 'myapp.change_model')

    Raises:
        AssertionError: If permission is not required

    Example:
        >>> assert_view_has_permission(
        ...     MyEditView,
        ...     'myapp.change_model'
        ... )
    """
    info = get_view_permissions(view_class)
    if permission not in info["permissions"]:
        raise AssertionError(
            f"{view_class.__name__} does not require permission '{permission}'. "
            f"Current permissions: {info['permissions']}"
        )


def assert_view_requires_login(view_class):
    """
    Assert that a view requires authentication.

    Args:
        view_class: The view class to check

    Raises:
        AssertionError: If view does not require login
    """
    if not issubclass(view_class, (LoginRequiredMixin, AccessMixin)):
        raise AssertionError(
            f"{view_class.__name__} does not require login. "
            f"Add LoginRequiredMixin or another auth mixin."
        )


# Backwards compatibility alias
assert_permissions_mixin_order = assert_mixin_order



from django import template
from django.template import Context
from django.template.base import token_kwargs
from loguru import logger

from . import quickforms

# from .slots.base import inject_node
from .slots.slot_node import do_define_slot, do_slot, parse_until

log = logger.debug

register = template.Library()


register.tag(name="slot.define")(do_define_slot)
register.tag(name="slot")(do_slot)



from django import template
from django.urls import resolve, reverse

register = template.Library()

quickform_kwargs = (
    "form_submit_label",
    "form_submit_button",
)


@register.inclusion_tag(
    "trim/quickform.html", takes_context=True, name="quickform.form"
)
def quickform_template(context, view_name, *args, **kwargs):
    """Apply a quickform within a form template, utilising the quickform as
    the thing to render.

        {% quickform.form 'associations:grouping-search-form'
            form_submit_button=True
            form_submit_label='Search'
            id='quicksearchform'
        %}

    Provide a `Form` instance as the first object to render a page form.

        {% quickform.form myform %}

    any option not in `quickform_kwargs` is applied to the form a node parameter.
    """
    form = view_name
    if isinstance(view_name, str):
        form = quickform(context, view_name, *args, **kwargs)

    extra = {"opts": {}, "trim_form": {}, "multipart": kwargs.get("multipart", True)}

    if getattr(form, "action_url", None) is None:
        kwpop = kwargs.pop("action_url", None)
        if kwpop is not None:
            form.action_url = kwpop

    for x, v in kwargs.items():
        # x, y for x, y in kwargs.items() if x in quickform_kwargs
        k = ["opts", "trim_form"][bool(x in quickform_kwargs)]
        extra[k][x] = v

    return {
        "trim_quickform_form": form,
        **extra,
    }


@register.simple_tag(takes_context=True, name="quickform")
def quickform(context, view_name, *url_args, **kwargs):
    """Build a form given the view_name and any `view_arguments`

    {% quickform 'associations:grouping-search-form' as grouping_search_form %}
    <form id='grouping_search_form'
        method='post'
        action='{{ grouping_search_form.action_url }}'>
        {{ grouping_search_form.as_ul }}
        {% csrf_token %}
        <button type='submit'>Submit</button>
    </form>

    """
    request = context.get("request")
    args = kwargs.get("view_args", ()) if len(url_args) == 0 else url_args

    rev = reverse(view_name, args=args)
    resolve_match = resolve(rev)
    action_url = kwargs.pop("action_url", rev)

    view = resolve_match.func.view_class
    initkwargs = resolve_match.func.view_initkwargs
    view.view_initkwargs = initkwargs

    instance = view(**initkwargs)
    instance.setup(request, *args, **kwargs)

    if not hasattr(instance, "get_form"):
        return
    form = instance.get_form()
    # Mutate the form to apply the action URL.
    form.action_url = action_url

    initial = kwargs.get("initial", None) or kwargs
    if hasattr(form, "initial") and (len(initial) > 0):
        form.initial.update(initial)
    return form



from django.contrib import admin

from trim import admin as tadmin

from . import models

# @tadmin.register(models.AccountEmail)
# class AccountEmail(admin.ModelAdmin):
#     list_display = (
#         'user',
#         'email_address',
#         'verified',
#         'created',
#         'updated',
#         )


tadmin.register_models(models, ignore=__name__)



from django.db import models
from wagtail.admin.panels import (
    FieldPanel,
    ObjectList,
    TabbedInterface,
    TitleFieldPanel,
)
from wagtail.api import APIField
from wagtail.models import Page

# from trim.wagtail import blocks as wbl
from trim.wagtail.streamfield import (  # FieldPanel,; APIField,; # as_fieldpanel_list,; # as_api_fields,
    prepared_streamfield,
)

# from wagtail import blocks


# class BlogPage(Page):
#     # field definitions omitted

#     content_panels = [
#         TitleFieldPanel('title', classname="title"),
#         FieldPanel('date'),
#         FieldPanel('body'),
#     ]


class StructuredPage(Page):
    author = models.CharField(max_length=255)
    # date = models.DateField("Post date")
    body = prepared_streamfield("default")

    styles_panels = [
        # FieldPanel('advert'),
        FieldPanel("author"),
        # InlinePanel('related_links', heading="Related links", label="Related link"),
    ]

    # content_panels = Page.content_panels + as_fieldpanel_list(
    #         'body',
    #         # 'date',
    #         # 'author',
    #     )
    content_panels = Page.content_panels + [
        # TitleFieldPanel('title', classname="title"),
        # FieldPanel('date'),
        FieldPanel("body"),
    ]

    api_fields = [
        # APIField('date'),
        APIField("body"),
        # APIField('author'),
    ]

    edit_handler = TabbedInterface(
        [
            ObjectList(content_panels, heading="Content"),
            ObjectList(styles_panels, heading="Styles"),
            ObjectList(Page.promote_panels, heading="Promote"),
        ]
    )

    class Meta:
        abstract = True



VIEWS = (
    ## crud
    "ListView",
    "CreateView",
    "UpdateView",
    "DeleteView",
    "DetailView",
    # 'DeleteViewCustomDeleteWarning',
    ## history
    "ArchiveIndexView",
    "DateDetailView",
    "DayArchiveView",
    "MonthArchiveView",
    "TodayArchiveView",
    "WeekArchiveView",
    "YearArchiveView",
    "RedirectView",
    "TemplateView",
    "FormView",
    "LoginView",
    "LogoutView",
    "PasswordChangeDoneView",
    "PasswordChangeView",
    "PasswordResetCompleteView",
    "PasswordResetConfirmView",
    "PasswordResetDoneView",
    "PasswordResetView",
)

## Order is importtant.
url_enforcements = (
    (
        (
            "update",
            "detail",
            "delete",
        ),
        "{name}/<str:pk>/",
    ),
    # ( ('archiveindex',), '', ),
    (
        ("todayarchive",),
        "today/",
    ),
    (
        ("datedetail",),
        "<int:year>/<str:month>/<int:day>/<int:pk>/",
    ),
    (
        ("dayarchive",),
        "<int:year>/<str:month>/<int:day>/",
    ),
    (
        ("montharchive",),
        "<int:year>/<str:month>/",
    ),
    (
        ("weekarchive",),
        "<int:year>/week/<int:week>/",
    ),
    (
        ("yeararchive",),
        "<int:year>/",
    ),
)

## Generate a Mapped name lise of the VIEWS, producing a list for direct
## name association:
# 'list':'ListView',
# 'create':'CreateView',
# 'update':'UpdateView',
# 'delete':'DeleteView',
# 'detail':'DetailView',
MAPPED_NAMES = {x[:-4].lower(): x for x in VIEWS}

## Write the reverse of the MAPPED_NAMES, for class to friendly map:
# ListView: list
MAPPED_CLASS = {y: x for x, y in MAPPED_NAMES.items()}


def crud():
    return (
        "list",
        "create",
        "update",
        "delete",
        "detail",
    )


def history():
    # Order is important
    return (
        "archiveindex",
        "todayarchive",
        "weekarchive",
        "datedetail",
        "dayarchive",
        "montharchive",
        "yeararchive",
    )


def tidy_enforcements(enforcements):
    url_parts = {}
    for names, enforced in enforcements:
        for n in names:
            url_parts[n] = enforced
    return url_parts


def render_defaults(names, parts):
    refs = {}
    res = {}
    for name, classname in names.items():
        refs["name"] = name
        refs["classname"] = classname
        live_part = parts.get(name, None)
        first = f"{name}/"
        part = ""
        if live_part is not None:
            part = live_part.format(**refs)
            first = ""
        res[name] = f"{first}{part}"
    return res


URL_PARTS = tidy_enforcements(url_enforcements)
URL_DEFAULTS = render_defaults(MAPPED_NAMES, URL_PARTS)


def get_url(name):
    return URL_DEFAULTS.get(name)


import inspect


def get_mapped_name(instance):
    mro = inspect.getmro(instance.__class__)
    for _c in mro:
        r = MAPPED_CLASS.get(_c.__name__)
        if r is not None:
            return r



"""Auth tooling, and convenient imports

Creating a login view:

    from trim import views

    class LoginView(views.LoginView):
        pass

URL Can be minimal:

    from trim.urls import paths_named
    from . import views

    app_name = 'userprofile'

    urlpatterns = [] + paths_named(views,
        login=('LoginView', '',),
    )


The template needs a form: `templates/userprofile/templates/registration/login.html`

    {% extends "home/base.html" %}
    {% load link quickforms %}

    {% block content.body %}
        {% quickform.form form %}
    {% endblock content.body %}


"""

from django.contrib.auth.mixins import UserPassesTestMixin
from django.contrib.auth.views import LoginView, LogoutView


def is_staff_or_admin(user):
    """
    Check if a user has staff or admin privileges.

    Args:
        user: A Django User object or user-like object with is_superuser
              and is_staff attributes.

    Returns:
        bool: True if the user is either a superuser or staff member,
              False otherwise.

    Example:

        >>> from django.contrib.auth.models import User
        >>> admin_user = User.objects.create(is_superuser=True)
        >>> is_staff_or_admin(admin_user)
        True
        >>> regular_user = User.objects.create(is_staff=False, is_superuser=False)
        >>> is_staff_or_admin(regular_user)
        False
    """
    return user.is_superuser or user.is_staff


class IsStaffMixin(UserPassesTestMixin):
    """
    A Test function mixin to restrict access to staff or admin users only.
    If the user is not active or not staff/admin, access is denied.

    Example:

        from trim.views import views

        class StaffOnlyView(views.IsStaffMixin, views.DetailView):
            model = models.SomeModel
            template_name = 'someapp/somemodel_detail.html'

    """

    def test_func(self):
        uu = self.request.user
        active_staff = uu.is_active and is_staff_or_admin(uu)
        return active_staff


class MissingField(Exception):
    pass


class UserOwnedMixin(UserPassesTestMixin):
    """
    Ensure the user owning the inner model `get_object()` is the requesting user.

        from trim.views import UserOwnedMixin
        class AddressDetailView(UserOwnedMixin, shorts.DetailView):
            model = models.Address
            user_field = 'creator'
            user_allow_staff = True
            template_name = 'locality/address_detail.html'
    """

    user_field = "user"
    user_allow_staff = False

    def test_func(self):
        uu = self.request.user
        request_pk = uu.pk
        model_user = getattr(self.get_object(), self.user_field)

        try:
            owns = model_user and (model_user.pk == request_pk)
        except AttributeError as err:
            raise MissingField(self.user_field) from err

        is_admin = is_staff_or_admin(uu) if self.user_allow_staff else False
        active_staff = uu.is_active and (owns or is_admin)

        return active_staff



"""
Unit tests for src/trim/models/live.py
"""

import unittest
from unittest.mock import patch

from trim.models import live
from trim.models.live import MagicModelApp, MagicModelModel


class TestMagicModelApp(unittest.TestCase):
    """Test the MagicModelApp class."""

    def setUp(self):
        self.magic_app = live

    @patch("django.apps.apps.get_model")
    def test_getattr_returns_magic_model_model(self, mock_get_model):
        # Setup: Access an app name attribute
        # Expected: Should return a MagicModelModel instance
        # Result: Check type
        model = live.baskets
        self.assertIsInstance(model, MagicModelModel)
        self.assertEqual(model.appname, "baskets")

        model = live.bilko.Florp
        mock_get_model.assert_called_once_with("bilko", "Florp")



from modelcluster.fields import ParentalKey
from wagtail.core import blocks
from wagtail.core.fields import StreamField
from wagtail.core.models import Orderable
from wagtail.images import get_image_model_string
from wagtail.images.blocks import ImageChooserBlock
from wagtail.snippets.models import register_snippet

from .. import panels
from .base import defaults
from .django import fk

register = register_snippet


def streamfield(*a, **kw):
    kw = defaults(a, kw, nil=True)
    return StreamField(*a, **kw)


def image_fk(*a, **kw):
    kw = defaults(a, kw, related_name="+")
    return fk(get_image_model_string(), *a, **kw)


image = image_fk



# Generated by Django 4.0.2 on 2022-08-25 23:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("account", "0003_alter_account_user"),
    ]

    operations = [
        migrations.CreateModel(
            name="ForgotPasswordRecord",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "email_address",
                    models.EmailField(blank=True, max_length=254, null=True),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
            ],
        ),
    ]



from django.contrib.auth.models import User
from django.contrib.auth.signals import user_logged_in
from django.db.models.signals import post_save
from django.dispatch import receiver

from . import models

# @receiver(user_logged_in)
# def post_login(sender, user, request, **kwargs):
#     print('Post Login event', request.session.get('cart_id'))


@receiver(post_save, sender=User)
def create_user_account(sender, instance, created, **kwargs):
    if created:
        models.Account.objects.create(user=instance)



# Generated by Django 4.0.2 on 2022-10-19 00:37

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("account", "0004_forgotpasswordrecord"),
    ]

    operations = [
        migrations.AddField(
            model_name="accountemail",
            name="user",
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



from django.db import models
from wagtail import blocks
from wagtail.admin.panels import FieldPanel
from wagtail.api import APIField
from wagtail.fields import StreamField
from wagtail.images.blocks import ImageChooserBlock

from trim.wagtail import blocks as wbl

FIELDS_CACHE = {"pre": ()}


def pre_install_global_block(name, class_):
    FIELDS_CACHE["pre"] += ((name, class_),)


def get_fields():
    FIELDS = {
        "default": [
            ("heading", HeadingBlock()),
            ("paragraph", RichTextBlock()),
            ("image", ImageChooserBlock()),
        ]
        + list(FIELDS_CACHE["pre"])
    }

    return FIELDS


class TabbedStructBlock(blocks.StructBlock):
    polyclasses = wbl.text(required=False, default="default", group="secondary")
    # dev_note = wbl.text(required=False, default='default', group='styles')

    def render_form_template(self):
        res = super().render_form_template()
        # context = self.get_form_context(
        #     self.get_default(), prefix="__PREFIX__", errors=None
        # )
        # return mark_safe(render_to_string(self.meta.form_template, context))
        return res

    class Meta:
        template = "blocks/default-content-block.html"
        form_template = "block_forms/tabbed-struct-block.html"
        # form_template = 'blocks/default-content-block.html'


class RichTextBlock(TabbedStructBlock):
    # polyclasses = blocks.CharBlock()
    # photo = ImageChooserBlock(required=False)
    content = wbl.richtext()
    # content = blocks.RichTextBlock()

    class Meta:
        icon = "text"
        form_classname = "richtext-block struct-block"
        # template = 'blocks/default-content-block.html'


class HeadingBlock(TabbedStructBlock):
    content = wbl.text()

    class Meta:
        icon = "text"
        form_classname = "title"
        template = "blocks/heading-block.html"


def as_fieldpanel_list(*items):
    """
    content_panels = Page.content_panels + [
        # FieldPanel('author'),
        # FieldPanel('date'),
        FieldPanel('body'),
    ]
    """
    res = ()
    for item in items:
        res += (FieldPanel(item),)
    return list(res)


def as_api_fields(*items):
    """
    # Export fields over the API
    api_fields = [
        # APIField('date'),
        APIField('body'),
        # APIField('author'),
    ]
    """
    res = ()
    for item in items:
        res += (APIField(item),)
    return list(res)


def prepared_streamfield(group="default", **kw):
    items = get_fields().get(group)
    return StreamField(items, **kw)



from django import template
from django.template.base import TemplateSyntaxError
from django.template.loader_tags import (
    BLOCK_CONTEXT_KEY,
    BlockContext,
    BlockNode,
    Node,
    Template,
    TextNode,
    construct_relative_path,
    do_extends,
)
from django.urls import resolve, reverse

from ..context import theming

register = template.Library()


@register.simple_tag(takes_context=True)
def theme_string(context):
    pass


from django.template.base import token_kwargs


@register.tag("theme")
def do_theme_extends(parser, token):
    # replace the token with the string tokeniser.
    bits = token.split_contents()
    extra = {}
    # bits[1] = construct_relative_path(parser.origin.template_name, bits[1])

    # Track the position of thr 'with' statement, to parse all string
    # bits before the position.
    with_position = None

    parts = ()
    for i, word in enumerate(bits):
        if word == "with":
            with_position = i
            extra = token_kwargs(bits[i + 1 :], parser)
            break
        if i == 0:
            continue
        # push bit into text ignoring 0
        parts += (word,)

    # Parse vars before i
    parent_name = parser.compile_filter(bits[1])
    # path = theming[parent_name.var]

    # if path is None:
    #     raise TemplateSyntaxError("'%s' Unknown theme name" % parent_name)

    template_vars = tuple(parser.compile_filter(x) for x in parts)
    # token.contents = ' '.join((bits[0], f'"{path}"',))

    # bits[1] = construct_relative_path(parser.origin.template_name, bits[1])
    # parent_name = parser.compile_filter(bits[1])

    nodelist = parser.parse()
    if nodelist.get_nodes_by_type(ThemeExtendsNode):
        raise TemplateSyntaxError(
            "'%s' cannot appear more than once in the same template" % bits[0]
        )

    return ThemeExtendsNode(
        nodelist, template_vars, extra_context=extra, origin=parser.origin
    )


class ThemeExtendsNode(Node):
    must_be_first = True
    context_key = "extends_context"

    def __init__(
        self, nodelist, template_vars, template_dirs=None, extra_context=None, **kw
    ):
        self.nodelist = nodelist
        self.parent_name = None
        self.template_vars = template_vars
        self.template_dirs = template_dirs
        self.blocks = {n.name: n for n in nodelist.get_nodes_by_type(BlockNode)}
        self.extra_context = extra_context or {}
        self.kwargs = kw or {}

    def get_parent_token(self):
        if self.parent_name:
            return self.parent_name.token
        return tuple(x.token for x in self.template_vars)

    def __repr__(self):
        return "<%s: extends %s>" % (self.__class__.__name__, self.get_parent_token())

    def find_template(self, template_name, context):
        """
        This is a wrapper around engine.find_template(). A history is kept in
        the render_context attribute between successive extends calls and
        passed as the skip argument. This enables extends to work recursively
        without extending the same template twice.
        """
        history = context.render_context.setdefault(
            self.context_key,
            [self.origin],
        )
        template, origin = context.template.engine.find_template(
            template_name,
            skip=history,
        )
        history.append(origin)
        return template

    def resolve_parent(self, context):
        if self.parent_name:
            return self.parent_name.resolve(context)
        return self.resolve_theme_parent(context)

    def resolve_theme_parent(self, context):
        """Resolve the synonym of the `parent_name` for the _theming_
        multi string"""
        # resolve_buts,
        res = ()
        for template_var in self.template_vars:
            rtv = template_var.resolve(context)
            res += (rtv,)

        # Provide the parts into the theming,
        position_bit = theming.resolve_theme_parent(res, self.origin, context)
        path = construct_relative_path(self.origin.template_name, position_bit)
        return path
        # path = theming[parent_name.var]

    def get_parent(self, context):
        parent = self.resolve_parent(context)
        if not parent:
            error_msg = "Invalid template name in 'extends' tag: %r." % parent
            if self.parent_name.filters or isinstance(self.parent_name.var, Variable):
                error_msg += (
                    " Got this from the '%s' variable." % self.parent_name.token
                )
            raise TemplateSyntaxError(error_msg)
        if isinstance(parent, Template):
            # parent is a django.template.Template
            return parent
        if isinstance(getattr(parent, "template", None), Template):
            # parent is a django.template.backends.django.Template
            return parent.template
        return self.find_template(parent, context)

    def render(self, context):
        compiled_parent = self.get_parent(context)

        if BLOCK_CONTEXT_KEY not in context.render_context:
            context.render_context[BLOCK_CONTEXT_KEY] = BlockContext()
        block_context = context.render_context[BLOCK_CONTEXT_KEY]

        # Add the block nodes from this node to the block context
        block_context.add_blocks(self.blocks)

        # If this block's parent doesn't have an extends node it is the root,
        # and its block nodes also need to be added to the block context.
        for node in compiled_parent.nodelist:
            # The ThemeExtendsNode has to be the first non-text node.
            if not isinstance(node, TextNode):
                if not isinstance(node, ThemeExtendsNode):
                    blocks = {
                        n.name: n
                        for n in compiled_parent.nodelist.get_nodes_by_type(BlockNode)
                    }
                    block_context.add_blocks(blocks)
                break

        extra = self.extra_context

        # Call Template._render explicitly so the parser context stays
        # the same.
        with context.render_context.push_state(compiled_parent, isolated_context=False):
            extra = {key: val.resolve(context) for key, val in extra.items()}
            print("Extra", extra)
            context.push(extra)
            return compiled_parent._render(context)


@register.filter(name="attr")
def get_attr(value, arg):
    """Return the _attribute_ of the given object, such as a model field from a
    model instance.

        {{ object|attr:field }}

    Example:

        {% load theming %}

        {% with field="field_name" %}
        The value of {{ field }} for model "{{ object }}" == {{ object|attr:field }}
        {% endwith %}

    """
    return getattr(value, arg)



from pathlib import Path

from django.template.base import TemplateSyntaxError

# from django.conf import settings
from . import config

## The range of template Theming classes for the {{ theme }} template tags.
# the 'default' theming is applied to the module `theming` instance.
themers = {}
theming = None


class NoThemingTemplateFound(Exception):
    pass


def find_template(mappings, name, root=None, version=None):
    map_filename = mappings.get(name)
    if map_filename is None:
        raise NoThemingTemplateFound(name)

    root = config.theme_option("root", root, "./")
    version = config.theme_option("version", version, "")

    if map_filename.startswith("/"):
        # No rooting - the template name is literal
        # /templates/[map_filename]
        return str(Path("") / version / map_filename[1:])
    return str(Path(root) / version / map_filename)


class Theming(object):

    def __init__(self, theme_map=None, key=None):
        self.theme_map = theme_map or {}
        self.key = key

    def get_theme_map(self):
        if callable(self.theme_map):
            return self.theme_map()
        return self.theme_map

    def _resolve(self, name=None):
        try:
            return find_template(self.get_theme_map(), name)
        except NoThemingTemplateFound:
            raise TemplateSyntaxError(f'theming name "{name}" did not resolve')

    def _template_resolve(self, name=None):
        res = self._resolve(name=name)
        if res is None:
            raise TemplateSyntaxError(f'theming name "{name}" did not resolve')
        return res

    def resolve_theme_parent(self, bits, origin, context):
        key = ".".join(bits)
        return self._template_resolve(name=key)

    def _loop_response(self, k):
        return self._resolve(k)

    def __getattr__(self, k):
        return self._loop_response(k)

    def __getitem__(self, k):
        return self._loop_response(k)

    def __str__(self):
        """Called by the template when applied to the view."""
        return self._template_resolve(name=self.key)


def build_default_themer():
    global theming
    theming = themers["default"] = Theming(config.get_theme_map)


def magic_strings(request):
    return {"theme": themers["default"]}



"""
Test trim.context module.

Tests the context processor functionality for Django templates.
"""

from django.http import HttpResponse
from django.test import RequestFactory, TestCase, override_settings
from django.urls import include, path


def dummy_view(request):
    """Dummy view for URL testing."""
    return HttpResponse("OK")


# Test URL patterns with app namespace
test_urlpatterns = [
    path(
        "test-app/",
        include(
            (
                [
                    path("page/", dummy_view, name="test_page"),
                ],
                "testapp",
            )
        ),
    ),
    path("no-app/", dummy_view, name="no_app_page"),
]


@override_settings(ROOT_URLCONF="tests.test_context")
class AppNameContextProcessorTest(TestCase):
    """Test the appname context processor."""

    def setUp(self):
        """Set up test fixtures."""
        self.factory = RequestFactory()

    def test_appname_with_app_namespace(self):
        """Test appname returns app_name when URL has an app namespace."""
        from trim.context import appname

        request = self.factory.get("/test-app/page/")
        request.path = "/test-app/page/"

        result = appname(request)
        self.assertIsInstance(result, dict)
        self.assertIn("appname", result)
        self.assertEqual(result["appname"], "testapp")

    def test_appname_without_app_namespace(self):
        """Test appname returns empty string when URL has no app namespace."""
        from trim.context import appname

        request = self.factory.get("/no-app/")
        request.path = "/no-app/"

        result = appname(request)
        self.assertIsInstance(result, dict)
        self.assertIn("appname", result)
        self.assertEqual(result["appname"], "")

    def test_appname_returns_dict(self):
        """Test that appname always returns a dictionary."""
        from trim.context import appname

        request = self.factory.get("/test-app/page/")
        request.path = "/test-app/page/"

        result = appname(request)
        self.assertIsInstance(result, dict)
        self.assertEqual(len(result), 1)


# Export urlpatterns for override_settings to work
urlpatterns = test_urlpatterns



from django import forms

from trim.forms import fields


def file_upload_loc(instance, filename):
    # file will be uploaded to MEDIA_ROOT/user_<id>/<filename>
    return "user_{0}/{1}".format("uploads/", filename)
    # return "user_{0}/{1}".format(instance.user.id, filename)


"""
1. serve a form with all the fields.
2. on post, edit the form in js
2. Post js file reader responses to the chunk form
    Each chunk response returns an update result
"""


class FileForm(forms.Form):
    """A Meta form is served in parallel to the file upload
    The meta should receive an upload ID, stored with the
    asset to the temp location until required.
    """

    filename = fields.chars(max_length=255, required=False)
    file = fields.file(required=False)
    # files = fields.files(required=False)

    byte_size = fields.hidden(fields.int(required=False))
    filepath = fields.hidden(fields.chars(max_length=255, required=False))
    filetype = fields.hidden(fields.chars(max_length=255, required=False))

    # class Meta:
    #     fields = ('query',)


class FilesForm(forms.Form):
    """A Meta form is served in parallel to the file upload
    The meta should receive an upload ID, stored with the
    asset to the temp location until required.
    """

    filename = fields.chars(max_length=255, required=False)
    # file = fields.file(required=False)
    files = fields.files(required=False)

    byte_size = fields.hidden(fields.int(required=False))
    filepath = fields.hidden(fields.chars(max_length=255, required=False))
    filetype = fields.hidden(fields.chars(max_length=255, required=False))

    # class Meta:
    #     fields = ('query',)


class FileChunkForm(forms.Form):
    """The JS chunking serves the FilesForm to receive additional
    parameters for this form.

    The JS slices the file into chunks, posting many chunks to this endpoint.
    """

    # chunk_id = forms.CharField()

    # The file_uuid is populated by the response from the FilesForm
    # Used to track a single upload
    file_uuid = forms.CharField()
    # The numerical index of the chunk - managed by the JS
    chunk_index = fields.integer()
    # The 'file' expecting bytes for a partial file
    filepart = fields.file()

    # class Meta:
    #     fields = ('query',)


class MergeConfirmForm(forms.Form):
    confirm_merge = fields.bool_true()
    delete_cache = fields.bool_true()



import string
import unittest

from trim.rand import rand_str


class RandTestCase(unittest.TestCase):
    """Test case for the trim rand module."""

    def test_rand_str_length(self):
        """Test that rand_str generates a string of the correct length."""
        length = 10
        result = rand_str(length)
        self.assertEqual(len(result), length)

    def test_rand_str_content(self):
        """Test that rand_str generates a string with valid characters."""
        result = rand_str()
        valid_chars = set(string.ascii_uppercase + string.digits)
        self.assertTrue(all(c in valid_chars for c in result))


class TestRandStr(unittest.TestCase):
    def setUp(self):
        # Setup: Prepare any required variables or state
        self.length = 6

    def test_rand_str_default_length(self):
        """Test rand_str returns a string of default length."""
        # Expected: String of length 6
        # Result: Call rand_str and check length
        result = rand_str()
        self.assertEqual(len(result), self.length)

    def test_rand_str_custom_length(self):
        """Test rand_str returns a string of specified length."""
        # Expected: String of custom length
        # Result: Call rand_str with custom length and check
        custom_length = 8
        result = rand_str(custom_length)
        self.assertEqual(len(result), custom_length)

    def test_rand_str_characters(self):
        """Test rand_str only uses uppercase letters and digits."""
        # Expected: All characters in result are valid
        # Result: Call rand_str and check character set
        result = rand_str()
        valid_chars = set(string.ascii_uppercase + string.digits)
        self.assertTrue(all(c in valid_chars for c in result))



from django.http import Http404

# from django.utils.translation import ugettext_lazy  as _
from django.utils.translation import gettext as _

from trim import views


class ProfileNewAccount(views.TemplateView):
    template_name = "account/new-profile.html"


class ProfileView(views.TemplateView):
    template_name = "account/profile.html"


class ProfileUpdateView(views.UpdateView):
    success_url = views.reverse_lazy("account:profile")
    template_name = "account/profile_form.html"
    model = views.get_user_model()
    fields = (
        "first_name",
        "last_name",
        # 'username',
    )

    def get_object(self, queryset=None):
        """
        Return the object the view is displaying.
        Require `self.queryset` and a `pk` or `slug` argument in the URLconf.
        Subclasses can override this to return any object.
        """
        # Use a custom queryset if provided; this is required for subclasses
        # like DateDetailView
        if queryset is None:
            queryset = self.get_queryset()
        # Next, try looking up by primary key.
        pk = self.kwargs.get(self.pk_url_kwarg)
        slug = self.kwargs.get(self.slug_url_kwarg)
        if pk is not None:
            queryset = queryset.filter(pk=pk)
        queryset = queryset.filter(pk=self.request.user.pk)

        try:
            # Get the single item from the filtered queryset
            obj = queryset.get()
        except queryset.model.DoesNotExist as err:
            d = {"verbose_name": queryset.model._meta.verbose_name}
            msg = _("No %(verbose_name)s found matching the query") % d
            raise Http404(msg) from err
        return obj


class ProfileUsernameUpdateView(ProfileUpdateView):
    fields = ("username",)


class ProfileEmailUpdateView(ProfileUpdateView):
    fields = ("email",)



from django.http import Http404

from trim import views

from .. import forms, models
from .profile import ProfileUpdateView

try:
    from mailroom import mail

    has_mail = True
except ImportError:
    print("mailroom app is not installed.")
    has_mail = False

from trim.urls import absolute_reverse


# class VerifiedEmailUpdateView(trim.):
class VerifiedEmailUpdateView(ProfileUpdateView):
    """Upon save generate a token and send an email
    When the user verifys, set the user primary to the new record.

    on update create change record
    send email to primary, and secondary
    Upon secondary accept, change the primary email.

    The flow is tokenized.,
    """

    success_url = views.reverse_lazy("account:profile")
    model = models.AccountEmail
    # model = views.get_user_model()
    fields = ("email_address",)

    def get_initial(self):
        return {
            "user": self.request.user,
            "email_address": self.request.user.email,
        }

    def form_valid(self, form):
        """If the form is valid, dont save the associated model."""
        self.object = form.save()
        self.send_email()
        return super().form_valid(form)

    def send_email(self):
        user_emails = self.request.user.email
        url = absolute_reverse(self.request, "account:verify_email_token")
        tokened_url = absolute_reverse(
            self.request, "account:verify_email_token", self.object.uuid_token
        )
        if has_mail:
            mail.template_send(
                "verify_email_change",
                user_emails,
                user=self.request.user,
                accountemail=self.object,
                url=url,
                tokened_url=tokened_url,
            )
        # new_email = self.object.email_address
        # mail.template_send('confirm_email_ownership', new_email,
        #     user=self.request.user,
        #     accountemail=self.object,
        #     url=url)

    def get_object(self, queryset=None):
        if queryset is None:
            return self.new_model()
        try:
            return super().get_object(queryset)
        except (AttributeError, models.AccountEmail.DoesNotExist, Http404):
            return self.new_model()

    def new_model(self):
        m, c = self.model.objects.get_or_create(
            user=self.request.user,
            verified=None,
        )
        return m


class VerifyEmailTokenView(views.FormView):
    """A user may enter a GET form withthe token in the header."""

    success_url = views.reverse_lazy("account:profile")
    model = models.AccountEmail
    form_class = forms.EmailChangeToken
    template_name = "account/form.html"
    query_pk_and_slug = True
    slug_url_kwarg = "uuid"
    slug_field = "uuid_token"

    def get_initial(self):
        r = self.initial.copy()
        r["token"] = self.kwargs.get("uuid")
        return r

    def get(self, request, *args, **kwargs):
        """
        Handle POST requests: instantiate a form instance with the passed
        POST variables and then check if it's valid.
        """
        form = self.get_form()
        if form.is_valid():
            return self.form_valid(form)
        return super().get(request, *args, **kwargs)

    def form_valid(self, form):
        """Test the form token against the verfied user and form."""
        data = form.cleaned_data
        token = data["token"]
        user = self.request.user
        try:
            ae = self.model.objects.get(uuid_token=token)
        except self.model.DoesNotExist:
            # inject error here.
            print("Cannot find model of", token)
            return self.form_invalid(form)

        if ae.user.id != user.id:
            # inject error here.
            return self.form_invalid(form)

        ae.verified = now()
        user.email = ae.email_address
        user.save()
        ae.save()

        return super().form_valid(form)


from django.utils.timezone import now



"""

Experiments:
    res = self.run_poll_command(command)

    res = self.run_command(command)

    subcall_stream(command)

    read_one_stream_command(command)

    gen = self.run_command2(command)
    for line in iter(gen):
        print(line, end='')

Raw method:
    print(res)
    res = subprocess.run(command, shell=True, capture_output=True)
    print('  code: ', res.returncode)
    out = clean(res.stdout)
    err = clean(res.stderr)
    print('  out:  ', out)
    print('  error:', err)
    print(' ', dir(res))

"""

import subprocess
import sys

import pkg_resources


def main():
    print("Inject")


__requires__ = "trim"
import sys

from pkg_resources import load_entry_point

if __name__ == "__main__":
    sys.exit(load_entry_point("django-trim==0.1", "console_scripts", "trim")())


def inj():
    # Create the fake entry point definition
    ep = pkg_resources.EntryPoint.parse("dummy = dummy_module:DummyPlugin")

    # Create a fake distribution to insert into the global working_set
    d = pkg_resources.Distribution()

    # Add the mapping to the fake EntryPoint
    d._ep_map = {"console_scripts": {"dummy": ep}}

    # Add the fake distribution to the global working_set
    pkg_resources.working_set.add(d, "dummy")


def test_entry_point():
    distribution = pkg_resources.Distribution(__file__)
    entry_point = pkg_resources.EntryPoint.parse(
        "plugin1=plugins.plugin1:plugin1_class", dist=distribution
    )
    distribution._ep_map = {"my_project.plugins": {"plugin1": entry_point}}
    pkg_resources.working_set.add(distribution)


def subcall_stream(cmd, fail_on_error=True):
    # Run a shell command, streaming output to STDOUT in real time
    # Expects a list style command, e.g. `["docker", "pull", "ubuntu"]`
    p = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        universal_newlines=True,
    )
    for line in p.stdout:
        sys.stdout.write(line)
    p.wait()

    return exit_code


def read_one_stream_command(command, show=True):
    process = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        ## not required when processing by read(1)
        # bufsize=1,
        ## Adding universal newlines will evoke _str_ from the `out`
        # universal_newlines=True,
    )
    res = ""
    byte_count = 0
    count = 0
    while True:
        out = process.stdout.read(1)
        if out == "" and process.poll() != None:
            break
        byte_count += 1

        if byte_count > 75_000:
            break

        if len(out) > 0:
            t = out.decode("utf")
            res += t
            if show:
                sys.stdout.write(t)
                sys.stdout.flush()
        else:
            count += 1
            if count % 20 == 0:
                print("blanked.")
                break
    # process.wait()
    return res


def run_poll_command(command):
    process = subprocess.Popen(command, stdout=subprocess.PIPE)
    while True:
        output = process.stdout.readline()
        if output == "" and process.poll() is not None:
            break
        if output:
            print(output.strip())
    rc = process.poll()
    return rc


def run_command(command):
    try:

        process = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            shell=True,
            encoding="utf8",
        )
        while True:
            inner = True
            pack = ""
            while inner:
                char = process.stdout.readline(1)
                pack += char
                if char == "\n":
                    print(pack, end="")
                if char is None:
                    print(pack)
                inner = False
            print(pack, end="")
            output = process.stdout.readline(64)
            if output == "" and process.poll() is not None:
                print("no output")
                break
            if output:
                print(output.strip())
        rc = process.poll()
        return rc
    except KeyboardInterrupt:
        # process.terminate()
        print("KeyboardInterrupt")


def run_command2(cmd):
    popen = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        # stdin=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
        shell=True,
        encoding="utf8",
        bufsize=1,
    )
    stdout = popen.stdout
    for stdout_line in iter(stdout.readline, ""):
        yield stdout_line
        stdout.flush()

    stdout.flush()
    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, cmd)
    return return_code


if __name__ == "__main__":
    main()



"""
Dedicated test suite for paths_dict function.
This function is complex enough to warrant its own test file.
"""

from unittest import TestCase
from unittest.mock import MagicMock, Mock, patch

from django.views.generic import CreateView, ListView

from trim import urls


class TestPathsDict(TestCase):
    """Test suite for paths_dict function."""

    @patch("trim.urls.paths")
    @patch("trim.urls.clean_str")
    def test_calls_clean_str_on_prefixes(self, mock_clean, mock_paths):
        """Test clean_str is called on all prefix parameters."""
        # Setup
        mock_clean.side_effect = lambda x: x or ""
        mock_paths.return_value = []
        mock_views = Mock()
        # Execute
        urls.paths_dict(
            mock_views,
            {},
            view_prefix="Product",
            url_pattern_prefix="api/",
            url_name_prefix="prod-",
        )
        # Assert
        self.assertEqual(mock_clean.call_count, 3)

    @patch("trim.urls.paths")
    def test_returns_result_from_paths(self, mock_paths):
        """Test paths_dict returns the result from paths()."""
        # Setup
        mock_paths.return_value = ["result"]
        mock_views = Mock()
        # Execute
        result = urls.paths_dict(mock_views, {})
        # Assert
        self.assertEqual(result, ["result"])

    @patch("trim.urls.paths")
    def test_patterns_defaults_to_empty_dict(self, mock_paths):
        """Test patterns parameter defaults to empty dict."""
        # Setup
        mock_paths.return_value = []
        mock_views = Mock()
        # Execute
        urls.paths_dict(mock_views, None)
        # Assert - should not raise error
        mock_paths.assert_called_once()

    @patch("trim.urls.paths")
    def test_merges_kwargs_into_patterns(self, mock_paths):
        """Test **kw params are merged into patterns dict."""
        # Setup
        mock_paths.return_value = []
        mock_views = Mock()
        mock_views.ListView = Mock()
        patterns = {"CreateView": ("create", "new/")}
        # Execute
        urls.paths_dict(mock_views, patterns, DetailView=("detail", "<pk>/"))
        # Assert - both patterns and kwargs should be processed
        call_arg = mock_paths.call_args[0][0]
        self.assertEqual(len(call_arg), 2)

    @patch("trim.urls.paths")
    def test_builds_class_name_with_view_prefix(self, mock_paths):
        """Test view_prefix is prepended to pattern keys."""
        # Setup
        mock_paths.return_value = []
        mock_views = Mock()
        mock_views.ProductListView = Mock()
        # Execute
        urls.paths_dict(mock_views, {"ListView": ("list", "")}, view_prefix="Product")
        # Assert - should call getattr with 'ProductListView'
        self.assertTrue(hasattr(mock_views, "ProductListView"))

    @patch("trim.urls.paths")
    def test_raises_attribute_error_when_view_missing(self, mock_paths):
        """Test AttributeError raised when view not found and ignore=False."""
        # Setup
        mock_paths.return_value = []
        mock_views = Mock(spec=[])
        # Execute & Assert
        with self.assertRaises(AttributeError):
            urls.paths_dict(
                mock_views,
                {"ListView": ("list", "")},
                ignore_missing_views=False,
            )

    @patch("trim.urls.paths")
    def test_continues_when_view_missing_and_ignore_true(self, mock_paths):
        """Test continues silently when view missing and ignore=True."""
        # Setup
        mock_paths.return_value = []
        mock_views = Mock(spec=[])
        # Execute
        result = urls.paths_dict(
            mock_views, {"ListView": ("list", "")}, ignore_missing_views=True
        )
        # Assert - should not raise, returns empty result
        self.assertEqual(result, [])

    @patch("trim.urls.paths")
    @patch("trim.urls.inspect.getmro")
    @patch("trim.urls.MAPPED_CLASS", {"ListView": "list"})
    def test_converts_string_url_to_tuple(self, mock_getmro, mock_paths):
        """Test string URL gets converted to tuple with MAPPED_CLASS lookup."""
        # Setup
        mock_paths.return_value = []
        mock_view = Mock()
        mock_getmro.return_value = (mock_view, ListView)
        mock_views = Mock()
        mock_views.ListView = mock_view
        # Execute
        urls.paths_dict(mock_views, {"ListView": ""})
        # Assert
        call_arg = mock_paths.call_args[0][0]
        self.assertEqual(len(call_arg), 1)
        fname, entry = call_arg[0]
        self.assertEqual(entry[0], "")  # URL

    @patch("trim.urls.paths")
    def test_unpacks_solution_tuple(self, mock_paths):
        """Test solution tuple is unpacked into path_name, url, extra."""
        # Setup
        mock_paths.return_value = []
        mock_view = Mock()
        mock_views = Mock()
        mock_views.ListView = mock_view
        # Execute
        urls.paths_dict(mock_views, {"ListView": ("list", "items/", {"extra": "data"})})
        # Assert
        call_arg = mock_paths.call_args[0][0]
        fname, entry = call_arg[0]
        self.assertEqual(entry[0], "items/")  # URL
        self.assertEqual(entry[2], {"extra": "data"})  # extra

    @patch("trim.urls.paths")
    def test_applies_url_pattern_prefix(self, mock_paths):
        """Test url_pattern_prefix is prepended to URL."""
        # Setup
        mock_paths.return_value = []
        mock_view = Mock()
        mock_views = Mock()
        mock_views.ListView = mock_view
        # Execute
        urls.paths_dict(
            mock_views,
            {"ListView": ("list", "items/")},
            url_pattern_prefix="api/",
        )
        # Assert
        call_arg = mock_paths.call_args[0][0]
        fname, entry = call_arg[0]
        self.assertEqual(entry[0], "api/items/")

    @patch("trim.urls.paths")
    def test_applies_url_name_prefix(self, mock_paths):
        """Test url_name_prefix is prepended to path name."""
        # Setup
        mock_paths.return_value = []
        mock_view = Mock()
        mock_views = Mock()
        mock_views.ListView = mock_view
        # Execute
        urls.paths_dict(
            mock_views, {"ListView": ("list", "")}, url_name_prefix="product-"
        )
        # Assert
        call_arg = mock_paths.call_args[0][0]
        fname, entry = call_arg[0]
        self.assertEqual(fname, "product-list")

    @patch("trim.urls.paths")
    def test_handles_multiple_urls_as_tuple(self, mock_paths):
        """Test URL as tuple creates multiple entries."""
        # Setup
        mock_paths.return_value = []
        mock_view = Mock()
        mock_views = Mock()
        mock_views.ListView = mock_view
        # Execute
        urls.paths_dict(mock_views, {"ListView": ("list", ("", "items/"))})
        # Assert
        call_arg = mock_paths.call_args[0][0]
        self.assertEqual(len(call_arg), 2)  # Two URLs

    @patch("trim.urls.paths")
    def test_safe_prefix_adds_part_name_to_duplicate(self, mock_paths):
        """Test safe_prefix=True prepends part_name to avoid collisions."""
        # Setup
        mock_paths.return_value = []
        mock_view = Mock()
        mock_views = Mock()
        mock_views.ListView = mock_view
        # Execute
        urls.paths_dict(mock_views, {"ListView": ("list", "")}, safe_prefix=True)
        # Assert
        call_arg = mock_paths.call_args[0][0]
        fname, entry = call_arg[0]
        self.assertTrue(fname.startswith("listview-"))

    @patch("trim.urls.paths")
    def test_forwards_extra_params_to_paths(self, mock_paths):
        """Test extra params in solution are forwarded."""
        # Setup
        mock_paths.return_value = []
        mock_view = Mock()
        mock_views = Mock()
        mock_views.ListView = mock_view
        extra_dict = {"permission": "view"}
        # Execute
        urls.paths_dict(mock_views, {"ListView": ("list", "", extra_dict)})
        # Assert
        call_arg = mock_paths.call_args[0][0]
        fname, entry = call_arg[0]
        self.assertIn(extra_dict, entry)



# from django.db import models

# Create your models here.
import inspect

from django.db import models as django_models


def is_model(name, unit):
    if name.startswith("__"):
        return False

    if inspect.isclass(unit):
        return django_models.Model in inspect.getmro(unit)

    return False


MODEL_CACHE = {}


def cache_known(*_models):
    for m in _models:
        MODEL_CACHE[m._meta.label] = MODEL_CACHE[m._meta.model_name] = m


def grab_models(_models, ignore=None):
    ignore = ignore or ()
    if isinstance(ignore, str):
        ignore = (ignore,)

    igs = ()
    for ig in ignore:
        if isinstance(ig, str) and ig.endswith(".admin"):
            # filter for any cached model, starting with the
            # same key
            a = ig.split(".")[0]
            a_dot = f"{a}."
            keep = {y for x, y in MODEL_CACHE.items() if x.startswith(a_dot)}
            igs += tuple(keep)
    ignore += igs

    items = ()

    for name in dir(_models):
        unit = getattr(_models, name)
        meta = getattr(unit, "_meta", None)
        if (name in ignore) or (unit in ignore) or (meta and (meta.abstract is True)):
            continue
        if is_model(name, unit):
            items += (unit,)
    return items


from django.apps import apps


def get_model(*a, **kw):
    return apps.get_model(*a, **kw)



from django.core import serializers
from django.core.serializers.python import Serializer
from django.http import JsonResponse
from django.views.generic import DetailView, ListView, TemplateView


class JsonSerializer(Serializer):
    # pass

    def get_dump_object(self, obj):
        return {}

    # def end_object( self, obj ):
    #     self._current['id'] = obj._get_pk_val()
    #     self._current.update(self.get_dump_object(obj) or {})
    #     self.objects.append( self._current )


class JSONResponseMixin(object):
    """
    A mixin that can be used to render a JSON response.
    """

    def render_to_json_response(self, context, **response_kwargs):
        """
        Returns a JSON response, transforming 'context' to make the payload.
        """
        return JsonResponse(
            context,
        )

    def get_data(self):
        """
        Returns an object that will be serialized as JSON by json.dumps().
        """
        return {}


class JsonView(JSONResponseMixin, TemplateView):
    """

    A JsonView is a JSONResponseMixin with TemplateView.

    This view serializes data and returns it as a JSON response.
    The data can be wrapped in an object with a property name specified by the `prop` attribute,
    or returned directly if `prop` is None.

    Attributes:
        prop (str): The property name to wrap the serialized data in. If None,
                    the data is returned unwrapped. Defaults to "object".
    Methods:
        get_serialiser(): Returns a JsonSerializer instance for data serialization.
        get(request, *args, **kwargs): Handles GET requests and returns JSON response
                                       containing the serialized data.

    Example:

        class MyJsonView(JsonView):
            prop = "my_data"

            def get_data(self):
                return {"key": "value"}

    This will return a JSON response like:

        {"my_data": {"key": "value"}}

    With `prop` set to None, it would return:

        class MyJsonView(JsonView):
            prop = None
            ...

    Result:

        {"key": "value"}
    """

    prop = "object"

    def get_serialiser(self):
        serial_data = JsonSerializer()
        # serial_data.get_dump_object = self.get_dump_object
        return serial_data

    def get(self, request, *args, **kwargs):
        # serial = self.get_serialiser()
        result = self.get_data()
        # r = serial.serialize([result])
        data = {self.prop: result} if self.prop is not None else result

        return self.render_to_json_response(data, **kwargs)


class JSONListResponseMixin(object):
    fields = None

    def render_to_json_response(self, context, **response_kwargs):
        """
        Returns a JSON response, transforming 'context' to make the payload.
        """
        return JsonResponse(context, **response_kwargs)

    def get_dump_object(self, obj):
        keys = self.fields or "__all__"

        if keys == "__all__":
            r = ()
            fields = obj.keys() if isinstance(obj, dict) else obj._meta.fields
            for field in fields:
                r += (field.attname,)
            keys = r
        return {x: getattr(obj, x) for x in keys}

    def get_serialiser(self):
        serial_data = JsonSerializer()
        serial_data.get_dump_object = self.get_dump_object
        return serial_data

    def serialize_result(self, result):
        """Convert the query list or list of results for the outbound render.

        By default this uses the django serialiser but can be changed for a custom
        iteration:

            def serialize_result(self, result):
                # return a list of ids
                return [x.id for x in result]
        """
        serial = self.get_serialiser()
        return serial.serialize(result)


class JsonListView(JSONResponseMixin, JSONListResponseMixin, DetailView):
    model = None
    prop = "object_list"
    # Apply any extra dictionary data to append into the response dictionary
    # _after_ serialisation.
    response_extra = None

    def get_results(self):
        return self.model.objects.all()

    def get_response_extra(self, result):
        return {"count": len(result), **(self.response_extra or {})}

    def get(self, request, *args, **kwargs):
        result = self.get_results()
        data = {
            self.prop: self.serialize_result(result),
            **self.get_response_extra(result),
        }
        return self.render_to_json_response(data, **kwargs)


class JsonDetailView(JsonListView):
    prop = "object"

    def get_results(self):
        return self.model.objects.get(id=self.kwargs["pk"])

    def get(self, request, *args, **kwargs):
        return self.json_response(**kwargs)

    def json_response(self, obj=None, **kwargs):
        data = self.generate_response(obj)
        return self.render_to_json_response(data)

    def generate_response(self, obj=None):
        result = obj or self.get_results()
        # serial = self.get_serialiser()
        # r = serial.serialize([result])
        r = self.serialize_result([result])
        data = {self.prop: r[0]}
        return data



DEFAULT_NIL = True


def defaults(args, params, nil_sub=True, nil_key="nil", **kw):

    if nil_sub:
        # nil subtract, or substitution
        #  Nil. being blank+null - where nil=True
        if (nil_key in kw) or (nil_key in params):
            # Fetch the value without treating False as missing.
            val = kw.get(nil_key) if (nil_key in kw) else params.get(nil_key)

            if isinstance(val, bool):
                # if the given value is a single bool,
                # unpack and multiply into the count of
                # args for the blank_null func,
                val = [val] * 2
                kw.update(**blank_null(*val))

    for k, v in kw.items():
        if k == nil_key:
            continue
        params.setdefault(k, v)
    params.pop("nil", None)
    return params


def blank_null(b=True, n=True):
    return {"blank": b, "null": n}



from collections import defaultdict


class SlofTuple(tuple):
    """A tuple class with the additional `slice of` method or `slof(i)`
    returning `x[n] of x in self`.

        content += (('c', 'b'), ('g', 'r'),)
        content.slof(0)
        ('c', 'g',)

    """

    def slof(self, index):
        return self.__class__(x[index] for x in self)


class EasyPermissionString(object):

    def __init__(self, strs=None, index=0, positions=None):
        self.index = index
        self.positions = positions or defaultdict(set)
        self.strs = strs or SlofTuple()

    def __getattr__(self, k):
        return self.push(self.index, k)

    def push(self, index, *items):
        # strs = self.strs + tuple((index, x,) for x in items)
        copy = self.__class__(self.strs, index, self.positions)
        copy.positions[index].update(items)
        print("Copied", copy.positions)
        return copy

    @property
    def crud(self):
        return self.push(
            2,
            "add",
            "view",
            "change",
            "delete",
        )

    def __add__(self, other):
        if isinstance(other, str):
            return self.push(self.index, other)
        if isinstance(other, (tuple, list)):
            return self.push(self.index, *other)

        raise Exception("Cannot add type", type(other))
        return self

    def as_tuple(self):
        # Iterate the strs, split on positions in each,
        count = 2
        if len(self.positions) > 0:
            count = max(self.positions.keys()) + 1
        print("count", count)
        for i in range(count):
            try:
                self.positions[i].add(self.strs[i][1])
            except IndexError:
                pass
            # for index, key in self.strs:
            pos = self.positions[i].copy()
            print(" ", pos)

    def flat(self):
        return ".".join([x[1] for x in self.strs])

    def __str__(self):
        return self.flat()


props = EasyPermissionString()


def test():
    perms = props
    v = perms.stocks.StockCount.crud
    v.as_tuple()

    # assert v == ('stocks.add_stockcount',)


if __name__ == "__main__":
    test()



"""
{% timedelta cart.updated cart.created format="" %}
{{ cart.created|timesince:cart.updated }}
{% human_timedelta cart.updated cart.created format="" %}
"""

from datetime import timedelta

from django import template
from django.urls import reverse
from django.utils.translation import ngettext

register = template.Library()


@register.simple_tag(takes_context=False, name="timedelta")
def timedelta_tag(late_time, early_time, *targs, **kwargs):
    """Calculate the time difference between two datetime objects.

    Returns a Python timedelta object representing the duration between
    two datetime values. This is useful for time-based calculations and
    comparisons in templates.

    Args:
        late_time: Later datetime object (end time)
        early_time: Earlier datetime object (start time)
        *targs: Additional positional arguments (unused)
        **kwargs: Additional keyword arguments (unused)

    Returns:
        timedelta: Time difference between late_time and early_time

    Examples:
        {% load datetime %}

        {# Basic usage #}
        {% timedelta order.completed order.created %}
        {# Output: datetime.timedelta object (e.g., 2 days, 3:45:12) #}

        {# Use in comparisons #}
        {% timedelta cart.updated cart.created as duration %}
        {% if duration.days > 7 %}
            <p>Cart is over a week old!</p>
        {% endif %}

        {# Access timedelta properties #}
        {% timedelta event.end event.start as duration %}
        <p>Duration: {{ duration.days }} days, {{ duration.seconds }} seconds</p>
    """
    td = late_time - early_time
    return td


@register.simple_tag(takes_context=False, name="human_timedelta")
def str_timedelta_tag(late_time, early_time, *targs, **kwargs):
    """Calculate time difference and format it in human-readable text.

    Returns a localized, human-friendly string representation of the time
    difference between two datetime objects. Automatically handles years,
    days, hours, minutes, and seconds with proper pluralization.

    Args:
        late_time: Later datetime object (end time)
        early_time: Earlier datetime object (start time)
        *targs: Additional positional arguments (unused)
        **kwargs: Additional keyword arguments (unused)

    Returns:
        str: Human-readable time difference (e.g., "2 days 3 hours 45 minutes")

    Examples:
        {% load datetime %}

        {# Basic usage #}
        {% human_timedelta order.completed order.created %}
        {# Output: "2 days 3 hours 15 minutes" #}

        {# Show how long ago something was updated #}
        <p>Last updated: {% human_timedelta now cart.updated %} ago</p>
        {# Output: "Last updated: 3 hours 22 minutes ago" #}

        {# Display order processing time #}
        <div class="order-info">
            <span>Processing time: {% human_timedelta order.shipped order.created %}</span>
        </div>
        {# Output: "Processing time: 1 day 6 hours 30 minutes" #}

        {# Compare with Django's timesince filter #}
        <p>Django's timesince: {{ cart.created|timesince:cart.updated }}</p>
        <p>Trim's human_timedelta: {% human_timedelta cart.updated cart.created %}</p>
    """
    td = late_time - early_time
    return localize_timedelta(td)


def localize_timedelta(delta):
    """Convert a timedelta object to a human-readable localized string.

    Breaks down a timedelta into years, days, hours, and minutes (or seconds
    if less than a minute), formatting each component with proper pluralization
    using Django's ngettext for internationalization support.

    Args:
        delta (timedelta): Time duration to format

    Returns:
        str: Formatted string with space-separated time components
             (e.g., "1 year 2 days 3 hours 15 minutes")

    Note:
        - Years are approximated as 365 days (does not account for leap years)
        - Shows minutes if duration >= 1 minute, otherwise shows seconds
        - All components use proper singular/plural forms via ngettext
        - Automatically localizes based on Django's language settings

    Examples:
        >>> from datetime import timedelta
        >>> localize_timedelta(timedelta(days=400, seconds=7320))
        '1 year 35 days 2 hours 2 minutes'

        >>> localize_timedelta(timedelta(seconds=45))
        '45 seconds'

        >>> localize_timedelta(timedelta(days=1, hours=2, minutes=30))
        '1 day 2 hours 30 minutes'
    """
    ret = []
    num_years = int(delta.days / 365)
    if num_years > 0:
        delta -= timedelta(days=num_years * 365)
        ret.append(ngettext("%d year", "%d years", num_years) % num_years)

    if delta.days > 0:
        ret.append(ngettext("%d day", "%d days", delta.days) % delta.days)

    num_hours = int(delta.seconds / 3600)
    if num_hours > 0:
        delta -= timedelta(hours=num_hours)
        ret.append(ngettext("%d hour", "%d hours", num_hours) % num_hours)

    seconds = delta.seconds
    num_minutes = int(seconds / 60)
    if num_minutes > 0:
        ret.append(ngettext("%d minute", "%d minutes", num_minutes) % num_minutes)
    else:
        ret.append(ngettext("%d second", "%d seconds", seconds) % seconds)
    return " ".join(ret)
